define(
    [
         'underscore',
         'jquery',
         'backbone',
         'controllers/Base',
         'models/Base',
         'models/search/Job',
         'models/charting/Charting',
         'models/search/Report',
         'models/classicurl',
         'models/services/search/jobs/ResultJsonCols',
         'models/services/search/jobs/ResultJsonRows',
         'models/shared/TimeRange',
         'models/services/search/jobs/Result',
         'models/services/search/jobs/Summary',
         'models/services/search/jobs/Timeline',
         'models/shared/fetchdata/ResultsFetchData',
         'collections/search/SelectedFields',
         'util/general_utils',
         'util/splunkd_utils',
         'util/time',
         'splunk.util'
    ],
    function(
        _,
        $,
        Backbone,
        BaseController,
        BaseModel,
        SearchJobModel,
        ChartingModel,
        ReportModel,
        classicUrlModel,
        ResultJsonColsModel,
        ResultJsonRowsModel,
        TimeRangeModel,
        ResultModel,
        SummaryModel,
        TimelineModel,
        ResultsFetchDataModel,
        SelectedFieldsCollection,
        general_utils,
        splunkd_utils,
        time_utils,
        splunkUtil
    ){
        /**
         * REQUIRES:
         * 
         * @param {Object} options {
         *     model: {
         *         application: <models.Application>,
         *         appLocal: <models.services.AppLocal>,
         *         user: <models.services.admin.User>
         *     },
         *     collection: {
         *         times: <collections.services.data.ui.TimesV2>,
         *         roles: <collections.services.authorization.Roles>
         *     },
         *     deferreds: {
         *         appLocal: $.Deferred (can be from fetch()),
         *         user: $.Deferred (can be from fetch()),
         *         times: $.Deferred (can be from fetch()),
         *         roles: $.Deferred (can be from fetch())
         *     }
         * }
         * 
         * CREATES:
         *     model: {
         *         result: <models.services.search.job.ResultsV2>,
         *         summary: <model.services.search.job.SummaryV2>,
         *         timeline: <model.services.search.job.TimelineV2>,
         *         searchJob: <models.services.search.Job>,
         *         report: <models.Report>,
         *         reportPristine: <models.Report>,
         *         chartWorking: <models.Charting>,
         *         resultJsonCols: <models.services.search.jobs.ResultJsonCols>,
         *         resultJsonRows: <models.services.search.jobs.ResultJsonRows>,
         *         timeRange: <models.TimeRange>
         *     },
         *     collection: {
         *         selectedFields: <collections.SelectedFields>,
         *         workflowActions: <collections.services.data.ui.WorkflowActions>
         *     }
         *     this.autoPause
         * 
         * 
         */

        return BaseController.extend({
            initialize: function() {
                BaseController.prototype.initialize.apply(this, arguments);

                this.autoPause = false;

                this.ui_prefs_filter = [
                     "^display\..*"
                ];

                this.job_filter = [
                     "^display\..*",
                     "^dispatch\.earliest_time$",
                     "^dispatch\.latest_time$"
                ];

                this.report_filter = [
                    "^display\..*",
                    "^dispatch\.earliest_time$",
                    "^dispatch\.latest_time$",
                    "^search$"
                ];

                this.url_filter = [
                    "^display\..*",
                    "^earliest$",
                    "^latest$",
                    "^q$"
                ];
                
                //models
                this.model.searchJob = new SearchJobModel({}, {delay: SearchJobModel.JOB_MAX_POLLING_INTERVAL, processKeepAlive: true, keepAliveInterval: SearchJobModel.DEFAULT_KEEP_ALIVE_INTERVAL});
                this.model.metaDataJob = new SearchJobModel({}, {delay: SearchJobModel.DEFAULT_METADATA_POLLING_INTERVAL, keepAliveInterval: SearchJobModel.DEFAULT_KEEP_ALIVE_INTERVAL});
                if (!this.model.report) {
                    this.model.report = new ReportModel();
                    this.instantiatedReport = true;
                }
                this.model.reportPristine = new ReportModel();
                if (!this.model.timeRange) {
                    this.model.timeRange = new TimeRangeModel();
                    this.instantiatedTimeRange = true;
                }
                
                //set the defaults for the pagination and sorting
                this.model.report.entry.content.set({
                    'display.prefs.events.offset': "0",
                    'display.prefs.statistics.offset': "0",
                    'display.events.table.sortDirection': 'asc',
                    'display.statistics.sortDirection': 'asc'
                });

                this.model.chartWorking = new ChartingModel();
                this.model.resultJsonCols = new ResultJsonColsModel();
                this.model.resultJsonRows = new ResultJsonRowsModel();
                this.model.result = new ResultModel();
                this.model.metaDataResult = new ResultModel();
                this.model.summary = new SummaryModel();
                this.model.timeline = new TimelineModel();

                //for testing we need the this.classicUrl on the instance
                this.classicUrl = classicUrlModel;

                //collections
                this.collection.selectedFields = new SelectedFieldsCollection();

                //data bootstrap
                this.bootstrapDeferred = $.Deferred();
                
                $.when(this.options.deferreds.uiPrefs).then(function() {
                    this.classicUrl.fetch({
                        success: function(model, response){
                            this.autoPause = this.classicUrl.get("auto_pause");
                            
                            //remove values that should not be permalinked
                            this.classicUrl.set({
                                'display.events.timelineEarliestTime': undefined,
                                'display.events.timelineLatestTime': undefined,
                                'display.prefs.events.offset': undefined,
                                'display.prefs.statistics.offset': undefined,
                                'display.statistics.sortColumn': undefined,
                                'display.statistics.sortDirection': undefined,
                                'display.events.table.sortColumn': undefined,
                                'display.events.table.sortDirection': undefined
                            });
                            
                            this.syncFromClassicURL();
                        }.bind(this)
                    });
                }.bind(this));

                $.when(
                    this.options.deferreds.times,
                    this.options.deferreds.roles,
                    this.options.deferreds.appLocal,
                    this.options.deferreds.user,
                    this.options.deferreds.eventRenderers,
                    this.options.deferreds.workflowActions,
                    this.bootstrapDeferred
                ).then(function(){
                    this.initializeViewsAndEvents();

                    this.setReplacementDOM();
                    
                    // pass the report to the 'ready' event, the router sometimes needs to know its name
                    this.trigger('ready', this.model.report);
                }.bind(this));
            },

            /**
             * 
             * The following methods must be implemented 
             * 
             */
            initializeViewsAndEvents: function(){
                throw new Error("You must implement initializeViewsAndEvents");
            },
            setReplacementDOM: function() {
                //is this.replacementDOM = fooView.render().el;
                throw new Error("You must implement setReplacementDOM");
            },
            /**
             * end interface methods 
             */

            syncFromClassicURL: function() {
                var jobFetchDeferred = $.Deferred(),
                    jobCreationDeferred = $.Deferred(),
                    reportFetchDeferred = $.Deferred(),
                    timeRangeDeferred = $.Deferred(),
                    reportIdFromUrl = this.classicUrl.get('s'),
                    searchFromUrl = splunkUtil.stripLeadingSearchCommand(this.classicUrl.get('q')),
                    jobIdFromUrl = this.classicUrl.get('sid'),
                    attrsFromUrl = this.classicUrl.filterByWildcards(this.url_filter, { allowEmpty: true }),
                    uiPrefsOffFromUrl = this.classicUrl.get('uiprefsoff'),
                    searchIsNew = true,
                    attrsFromJob = {},
                    attrsFromUIPrefs = {},
                    searchFromReport, searchFromJob, search;
                
                //for testing purposes we will not use UI prefs if the url tells us to
                if (!uiPrefsOffFromUrl) {
                    attrsFromUIPrefs = this.model.uiPrefs.entry.content.filterByWildcards(this.job_filter);
                }

                //we need to mediate the legacy attrs from the URL to the report
                this.mediateClassicUrlAttrsToReport(attrsFromUrl);

                //if we don't have a report, a job, or a search string to load then we should show initial help
                if (!reportIdFromUrl && !jobIdFromUrl && !searchFromUrl) {
                    //merge the attrs from the url into the report
                    this.model.report.entry.content.set($.extend({}, attrsFromUIPrefs, attrsFromUrl));

                    //need to bootstrap the timerange for the timerangepicker
                    this.timeRangeBootstrap(timeRangeDeferred);

                    //we need to show the initial help page with the metadata job
                    $.when(timeRangeDeferred).then(function(){
                        this.startMetaDataSearch(this.bootstrapDeferred);
                    }.bind(this));
                    return;
                }

                this.reportBootstrap(reportFetchDeferred, jobCreationDeferred, reportIdFromUrl, searchFromUrl, jobIdFromUrl);
                this.jobBootstrap(jobFetchDeferred, jobCreationDeferred, reportIdFromUrl, searchFromUrl, jobIdFromUrl);

                $.when(reportFetchDeferred, jobFetchDeferred).then(function() {
                    searchFromReport = this.model.report.entry.content.get('search');
                    searchFromJob = splunkUtil.stripLeadingSearchCommand(this.model.searchJob.getSearch());
                    
                    //merge the attrs from the ui prefs into the report
                    if (this.model.report.isNew()) {
                        //the ui prefs should trump the report defaults    
                        this.model.report.entry.content.set($.extend({}, attrsFromUIPrefs));
                    } else {
                        //the values we get from the report should trump the ui pref values
                        var defaults = this.model.report.entry.content.toJSON();
                        _.defaults(defaults, attrsFromUIPrefs);
                        this.model.report.entry.content.set($.extend({}, defaults));
                    }

                    //merge the attrs from the job and the url into the report
                    if (this.model.searchJob.isNew()) {
                        if (searchFromUrl && ($.trim(searchFromUrl) !== $.trim(searchFromReport))) {
                            search = searchFromUrl;
                        } else {
                            search = searchFromReport;
                        }
                    } else {
                        if (searchFromUrl && ($.trim(searchFromUrl) !== $.trim(searchFromJob))) {
                            search = searchFromUrl;
                        } else {
                            searchIsNew = false;
                            search = searchFromJob;
                            
                            //get the attrs from the job for mediation
                            attrsFromJob = this.model.searchJob.entry.content.custom.filterByWildcards(this.job_filter);
                            attrsFromJob["display.page.search.mode"] = this.model.searchJob.getAdhocSearchMode();
                            attrsFromJob["dispatch.earliest_time"] = this.model.searchJob.getDispatchEarliestTimeOrAllTime();
                            attrsFromJob["dispatch.latest_time"] = this.model.searchJob.getDispatchLatestTimeOrAllTime();
                        }
                    }
                
                    if (searchIsNew) {
                        //merge the attrs from the url into the report
                        this.model.report.entry.content.set($.extend({}, attrsFromUrl, {search: search}));
                        this.startNewSearch(jobCreationDeferred, search, attrsFromUrl);
                    } else {
                        //merge the attrs from the url and the job into the report
                        this.model.report.entry.content.set($.extend({}, attrsFromUrl, attrsFromJob, {search: search}));
                        jobCreationDeferred.resolve();
                    }

                    this.populateSelectedFieldsFromReport();
                    this.timeRangeBootstrap(timeRangeDeferred);
                }.bind(this));

                $.when(jobCreationDeferred, timeRangeDeferred).then(function() {
                    if (!this.model.searchJob.isNew()){
                        if (!this.model.searchJob.isPreparing()) {
                            this.model.report.setDisplayType(this.model.searchJob.isReportSearch());
                            this.registerSearchJobFriends();
                        }
                        this.populateClassicUrlFromSearchJob();
                        this.model.searchJob.startPolling();                        
                    }
                    this.bootstrapDeferred.resolve();
                }.bind(this));
            },
            reportBootstrap: function(reportFetchDeferred, jobCreationDeferred, reportIdFromUrl, searchFromUrl, jobIdFromUrl) {
                if (reportIdFromUrl) {
                    this.model.report.set("id", this.applyModelFullPathToName(this.model.report, reportIdFromUrl));
                }
                
                //fetch the report no matter what
                //if we don't have an id then we need the defaults
                this.model.report.fetch({
                    data: {
                        app: this.model.application.get("app"),
                        owner: this.model.application.get("owner")
                    },
                    
                    success: function(model, response){
                        //set the report pristine before anything has been set on the report
                        this.model.reportPristine.setFromSplunkD(this.model.report.toSplunkD());
                        reportFetchDeferred.resolve();                            
                    }.bind(this),

                    error: function(model, response){
                        var errorResponse = response;

                        this.model.report.unset("id");
                        this.classicUrl.save({s: undefined}, {replaceState: true});

                        if (reportIdFromUrl && !searchFromUrl && !jobIdFromUrl){
                            //the user only provided a bad reportId
                            //we should initialize the UI with the bad report
                            this.model.report._onerror(this.model.report, response);
                            reportFetchDeferred.resolve();
                            jobCreationDeferred.resolve();
                        } else {
                            //fetch the report defaults and populate the report with them, and put the report into an error state
                            this.model.reportPristine.fetch({
                                success: function(model, response) {
                                    this.model.report.setFromSplunkD(this.model.reportPristine.toSplunkD());
                                    this.model.report.trigger("error", this.model.report, errorResponse);
                                    reportFetchDeferred.resolve();
                                }.bind(this),
                                error: function(model, response) {
                                    //this is extremely bad, it means we couldn't fetch _new
                                    reportFetchDeferred.resolve();
                                }.bind(this)
                            });
                        }
                    }.bind(this)
                });
            },
            jobBootstrap: function(jobFetchDeferred, jobCreationDeferred, reportIdFromUrl, searchFromUrl, jobIdFromUrl){
                if (jobIdFromUrl) {
                    //only fetch the job if we have an id
                    this.model.searchJob.set("id", jobIdFromUrl);
                    this.model.searchJob.fetch({
                        success: function(model, response){
                            jobFetchDeferred.resolve();
                        }.bind(this),

                        error: function(model, response){
                            this.classicUrl.save({sid: undefined}, {replaceState: true});
                            this.model.searchJob.unset("id");

                            if (jobIdFromUrl && !searchFromUrl && !reportIdFromUrl){

                                //the user only provided a bad jobId
                                //we should initialize the UI with the bad job
                                jobFetchDeferred.resolve();
                                jobCreationDeferred.resolve();
                            } else {
                                //save this error for when we make the job
                                this.jobBootstrapErrors = [
                                     splunkd_utils.createMessageObject(
                                         splunkd_utils.WARNING,
                                         _("The job could not be loaded. A new job has been started with the same search.").t()
                                     )
                                ];

                                //we have other searches to use
                                //don't broadcast an error
                                jobFetchDeferred.resolve();
                            }
                        }.bind(this)
                    });
                } else {
                    jobFetchDeferred.resolve();
                }
            },
            timeRangeBootstrap: function(timeRangeDeferred) {
                this.model.timeRange.save(
                    {
                        'earliest': this.model.report.entry.content.get('dispatch.earliest_time'),
                        'latest': this.model.report.entry.content.get('dispatch.latest_time')
                    },
                    {
                        validate: false,
                        success: function(model, response) {
                            timeRangeDeferred.resolve();
                        }.bind(this),
                        error: function(model, response) {
                            timeRangeDeferred.resolve();
                        }.bind(this)
                    }
                );
            },
            registerSearchJobFriends: function() {
                var adhocMode = (this.model.report.entry.content.get('display.page.search.mode') || "").toLowerCase(),
                    reportSearch = this.model.searchJob.isReportSearch(),
                    eventsId = this.model.searchJob.entry.links.get(SearchJobModel.EVENTS);

                if (reportSearch) {
                    this.model.searchJob.registerJobProgressLinksChild(SearchJobModel.RESULTS_PREVIEW, this.model.resultJsonCols, this.fetchResultJSONColumn, this);
                    this.model.searchJob.registerJobProgressLinksChild(SearchJobModel.RESULTS_PREVIEW, this.model.resultJsonRows, this.fetchResultJSONRows, this);
                }
                
                if ((reportSearch && adhocMode === splunkd_utils.VERBOSE) || (!reportSearch)) {
                    this.model.searchJob.registerJobProgressLinksChild(SearchJobModel.SUMMARY, this.model.summary, this.fetchSummary, this);
                    this.model.searchJob.registerJobProgressLinksChild(SearchJobModel.TIMELINE, this.model.timeline, this.fetchTimeline, this);

                    if (eventsId) {
                        this.model.result.set("id", eventsId);

                        if (this.model.searchJob.entry.content.get('isDone')) {
                            this.fetchResult();
                            return; //we do not want to bind to the job progress event 
                        }
                        
                        if (this.model.searchJob.entry.content.get('isPaused')) {
                            this.fetchResult();
                        }
                        
                        this.model.searchJob.on("jobProgress", function() {
                            var eventSorting = this.model.searchJob.entry.content.get('eventSorting'),
                                resultsLength = this.model.result.results.length,
                                resultsCount = parseInt(this.model.report.entry.content.get('display.prefs.events.count'), 10),
                                eventAvailableCountChanged = _.isNumber(parseInt(this.model.searchJob.entry.content.changedAttributes().eventAvailableCount, 10));

                            if(this.model.searchJob.isRealtime()) {
                                this.fetchResult();
                                return;
                            } 
                            
                            if((eventAvailableCountChanged || this.model.searchJob.entry.content.get('isDone')) && (eventSorting === "none")) { // | reverse
                                this.fetchResult();
                                return;
                            } 
                            
                            if (eventSorting === "desc" && (resultsLength < resultsCount)) {
                                this.fetchResult(); 
                                return;
                            }

                        }, this);
                    }
                }                
            },
            fetchResult: function() {
                var field_list = _(this.collection.selectedFields.pluck('name')).union(['_raw', '_time', '_audit', '_decoration', 'eventtype', 'linecount', '_fulllinecount']);
                if (this.model.searchJob.isRealtime()) {
                    field_list.push('_serial', 'splunk_server');
                }
                var options = {
                    data: {
                        offset: this.model.report.entry.content.get('display.prefs.events.offset'),
                        count: this.model.report.entry.content.get('display.prefs.events.count'),
                        earliest_time: this.model.report.entry.content.get('display.events.timelineEarliestTime'),
                        latest_time: this.model.report.entry.content.get('display.events.timelineLatestTime'),
                        segmentation:  ((this.model.report.entry.content.get('display.events.type') === 'table') ? 
                                        'none':
                                        this.model.report.entry.content.get('display.events.list.drilldown')),
                        max_lines: this.model.report.getNearestMaxlines(),
                        field_list: splunkUtil.fieldListToString(field_list),
                        truncation_mode: 'abstract'
                    },
                    sparseMode: true 
                };

                if(this.model.report.entry.content.get('display.events.type')==='table') {
                    if(!this.model.searchJob.entry.content.get('isRealTimeSearch')) { 
                        if(this.model.report.entry.content.get('display.events.table.sortColumn')) {
                            options.data.search = this.model.report.getSortingSearch();
                        }
                    } else {
                        this.model.report.entry.content.unset('display.events.table.sortColumn');
                    }
                }
                if (this.model.searchJob.entry.content.get('isRealTimeSearch')) {
                    options.data.offset = -(options.data.offset) - options.data.count;
                }
                this.model.result.safeFetch(options);
            },
            fetchResultJSONColumn: function() {
                if (this.model.searchJob.entry.content.get('isPreviewEnabled') || this.model.searchJob.entry.content.get('isDone')) {
                    this.model.resultJsonCols.safeFetch({
                        data: {
                            count: 1000,
                            output_mode: 'json_cols'
                        }
                    });
                }
            },
            fetchResultJSONRows: function() {
                if (this.model.searchJob.entry.content.get('isPreviewEnabled') || this.model.searchJob.entry.content.get('isDone')) {
                    var fetchDataModel = new ResultsFetchDataModel({
                        sortKey: this.model.report.entry.content.get("display.statistics.sortColumn"),
                        sortDirection: this.model.report.entry.content.get("display.statistics.sortDirection")
                    });
    
                    var data = $.extend(
                        fetchDataModel.toJSON(),
                        {
                            count: this.model.report.entry.content.get("display.prefs.statistics.count"),
                            offset: this.model.report.entry.content.get("display.prefs.statistics.offset")
                        }
                    );
    
                    this.model.resultJsonRows.safeFetch({
                        data: data
                    });
                }
            },
            fetchSummary: function() {
                this.model.summary.safeFetch({
                    data: {
                        min_freq: 0,
                        earliest_time: this.model.report.entry.content.get('display.events.timelineEarliestTime'),
                        latest_time: this.model.report.entry.content.get('display.events.timelineLatestTime')
                    }
                });
            },
            fetchTimeline: function() {
                var earliest_time = this.model.report.entry.content.get('display.events.timelineEarliestTime'),
                    latest_time = this.model.report.entry.content.get('display.events.timelineLatestTime');
                if (earliest_time && latest_time) {
                    this.model.timeline.safeFetch({
                        data: {
                            earliest_time: earliest_time,
                            latest_time: latest_time
                        }
                    });
                }
            },
            startNewSearch: function(jobCreationDeferred, search, reportMediatedAttrs) {
                var searchToStart;
                if (search){
                    searchToStart = new SearchJobModel();
                    if (reportMediatedAttrs) {
                        this.deleteOneTimeUseAttrs(reportMediatedAttrs);
                        searchToStart.entry.content.custom.set(reportMediatedAttrs);
                    }
                    //let the full model lifecycle take place before proxying
                    searchToStart.on('sync', function(model, response, options) {
                        var messages;
                        this.model.searchJob = new SearchJobModel({}, {
                            splunkDPayload: searchToStart.toSplunkD(),
                            delay: SearchJobModel.JOB_MAX_POLLING_INTERVAL,
                            processKeepAlive: true,
                            keepAliveInterval: SearchJobModel.DEFAULT_KEEP_ALIVE_INTERVAL
                        });
                        if (this.jobBootstrapErrors && this.jobBootstrapErrors.length > 0) {
                            messages = this.jobBootstrapErrors;
                        } else if (model.error.get('messages')) {
                            messages = model.error.get('messages');
                        }
                        if (messages) {
                            this.model.searchJob.error.set({
                                messages: messages
                            });
                        }
                        //clear auto_pause
                        this.classicUrl.unset("auto_pause");
                        searchToStart.off();
                        jobCreationDeferred.resolve();
                    }, this);
                    searchToStart.on('error', function(model, response, option) {
                        this.model.searchJob.trigger('error', this.model.searchJob, response);
                        searchToStart.off();
                        jobCreationDeferred.resolve();
                    }, this);
                    searchToStart.save({}, {
                        data: {
                            search: search,
                            earliest_time: this.model.report.entry.content.get('dispatch.earliest_time'),
                            latest_time: this.model.report.entry.content.get('dispatch.latest_time'),
                            auto_cancel: SearchJobModel.DEFAULT_AUTO_CANCEL,
                            auto_pause: this.autoPause ? SearchJobModel.DEFAULT_AUTO_PAUSE : undefined,
                            status_buckets: 300,
                            ui_dispatch_app: this.model.application.get('app'),
                            preview: this.model.report.entry.content.get('display.general.enablePreview'),
                            adhoc_search_level: this.model.report.entry.content.get('display.page.search.mode'),
                            app: this.model.application.get('app'),
                            owner: this.model.application.get('owner'),
                            indexedRealtime: this.model.report.entry.content.get('dispatch.indexedRealtime')
                        }
                    });
                } else {
                    this.jobBootstrapErrors = [
                        splunkd_utils.createMessageObject(
                            splunkd_utils.FATAL,
                            _('No search was specified.').t()
                        )
                    ];
                    jobCreationDeferred.resolve();
                }
            },
            registerMetaDataJobFriends: function() {
                this.model.metaDataJob.registerJobProgressLinksChild(SearchJobModel.RESULTS_PREVIEW, this.model.metaDataResult, this.fetchMetaDataResult, this);
            },
            fetchMetaDataResult: function() {
                var resultPreviewCount = this.model.metaDataJob.entry.content.get("resultPreviewCount");
                if (_.isNumber(resultPreviewCount) && (resultPreviewCount > 0)) {
                    this.model.metaDataResult.safeFetch({
                        data: {
                            count: 1,
                            search: "| stats sum(totalCount) as cnt, min(firstTime) as min, max(lastTime) as max"
                        }
                    });
                }
            },
            startMetaDataSearch: function(deferred) {
                var search = '| metadata type=sourcetypes | search totalCount > 0';

                this.model.metaDataJob.save({}, {
                    data: {
                        app: this.model.application.get("app"),
                        owner: this.model.application.get("owner"),
                        search: search,
                        preview: "true",
                        earliest_time: "rt",
                        latest_time: "rt",
                        auto_cancel: SearchJobModel.DEFAULT_AUTO_CANCEL,
                        max_count: 100000
                    },
                    success: function(model, response) {
                        deferred.resolve();
                        this.registerMetaDataJobFriends();
                        this.model.metaDataJob.startPolling();
                    }.bind(this),
                    error: function(model, response) {
                        deferred.resolve();
                    }.bind(this)
                });
            },
            populateClassicUrl: function(changedAttrs) {
                var changedAttrsModel, urlFilteredAttrs;
                
                if (changedAttrs && !_.isEmpty(changedAttrs)){
                    changedAttrs = $.extend(true, {}, changedAttrs);
                    changedAttrsModel = new BaseModel(changedAttrs);
                    urlFilteredAttrs = changedAttrsModel.filterByWildcards(this.report_filter, {allowEmpty: true});

                    //we have to mediate legacy attributes in the report back to the URL
                    this.mediateReportAttrsToClassicUrl(urlFilteredAttrs);

                    if (!_.isEmpty(urlFilteredAttrs)){
                        var search = this.model.report.entry.content.get('search');
                        if (search && (
                                _.has(urlFilteredAttrs,"q") || _.has(urlFilteredAttrs,"earliest") ||
                                _.has(urlFilteredAttrs,"latest") || _.has(urlFilteredAttrs,"display.page.search.mode")
                            )
                        ){                            
                            //unset the sid and oneTimeUserAttrs so that we get a new one when the page loads
                            urlFilteredAttrs['display.events.timelineEarliestTime'] = undefined;
                            urlFilteredAttrs['display.events.timelineLatestTime'] = undefined;
                            urlFilteredAttrs['display.prefs.events.offset'] = undefined;
                            urlFilteredAttrs['display.prefs.statistics.offset'] = undefined;
                            urlFilteredAttrs['display.statistics.sortColumn'] = undefined;
                            urlFilteredAttrs['display.statistics.sortDirection'] = undefined;
                            urlFilteredAttrs['display.events.table.sortColumn'] = undefined;
                            urlFilteredAttrs['display.events.table.sortDirection'] = undefined;
                            urlFilteredAttrs.sid = undefined;
                            urlFilteredAttrs.q = splunkUtil.addLeadingSearchCommand(search);

                            //this means we have changed the search, so push a back button state
                            this.classicUrl.save(urlFilteredAttrs, {
                                success: function(model, response){
                                    if (!this.classicUrl.hasChanged("q")){
                                        this.model.report.entry.content.trigger("enableSearchInput"); 
                                    }
                                }.bind(this),
                                trigger: true
                            });
                        } else {
                            //update the current back button state
                            this.classicUrl.save(urlFilteredAttrs, {
                                replaceState: true
                            });
                        }
                    }
                }
            },
            populateJob: function(jobPopulationDeferred, changedAttrs) {
                var changedAttrsModel, filteredAttrs, changedAttrsCopy;

                if (changedAttrs && !_.isEmpty(changedAttrs) &&
                        !this.model.searchJob.isNew() &&
                        this.model.searchJob.entry.acl.canWrite()){
                    changedAttrsCopy = $.extend(true, {}, changedAttrs);
                    changedAttrsModel = new BaseModel(changedAttrsCopy);
                    filteredAttrs = changedAttrsModel.filterByWildcards(this.report_filter, {allowEmpty: true});
                    if (!(
                            _.has(filteredAttrs,"search") || _.has(filteredAttrs,"dispatch.earliest_time") ||
                            _.has(filteredAttrs,"dispatch.latest_time") || _.has(filteredAttrs,"display.page.search.mode")
                    )){
                        this.deleteOneTimeUseAttrs(filteredAttrs);
                        if (!_.isEmpty(filteredAttrs)){
                            this.ensureTimeRangeFromReportInAttrs(filteredAttrs);  
                            this.model.searchJob.entry.content.custom.set(filteredAttrs);
                            this.model.searchJob.save({}, {
                                success: function(model, response){
                                    jobPopulationDeferred.resolve();
                                },
                                error: function(model, response){
                                    jobPopulationDeferred.resolve();
                                }
                            });
                            return;
                        }
                    }
                }
                jobPopulationDeferred.resolve();
            },
            populateUIPrefs: function(uiPrefsPopulationDeferred, changedAttrs) {
                var data = {},
                    changedAttrsModel, filteredAttrs, changedAttrsCopy;

                if (changedAttrs && !_.isEmpty(changedAttrs)) {
                    changedAttrsCopy = $.extend(true, {}, changedAttrs);
                    changedAttrsModel = new BaseModel(changedAttrsCopy);
                    filteredAttrs = changedAttrsModel.filterByWildcards(this.ui_prefs_filter, {allowEmpty: true});
                    this.deleteOneTimeUseAttrs(filteredAttrs);
                    
                    if (!_.isEmpty(filteredAttrs)){
                        if (this.model.uiPrefs.isNew()) {
                            data = {
                                app: this.model.application.get("app"),
                                owner: this.model.application.get("owner")
                            };
                        }

                        this.model.uiPrefs.entry.content.set(filteredAttrs);
                        this.model.uiPrefs.save({}, {
                            data: data,
                            success: function(model, response) {
                                uiPrefsPopulationDeferred.resolve();
                            }.bind(this),
                            error: function(model, response) {
                                uiPrefsPopulationDeferred.resolve();
                            }.bind(this)
                        });
                        return;
                    }
                }
                uiPrefsPopulationDeferred.resolve();
            },
            populateSelectedFieldsFromReport: function() {
                var models = [];
                //turn off the listener so that we don't have a circular reference
                this.collection.selectedFields.off('add remove reset', this.populateReportFromSelectedFields);

                //set fields into a collection (facilitates mgt)
                _.each(this.model.report.entry.content.toObject('display.events.fields') || [], function(value) {
                    models.push({name: value});
                });

                this.collection.selectedFields.reset(models);

                //turn the listener so that we don't have a circular reference
                this.collection.selectedFields.on('add remove reset', this.populateReportFromSelectedFields, this);
            },
            populateReportFromSelectedFields: function() {
                this.model.report.entry.content.set('display.events.fields', this.collection.selectedFields.valuesToJSONString());
            },
            populateClassicUrlFromSearchJob: function() {
                var attrsFromJob = this.model.searchJob.entry.content.custom.filterByWildcards(this.job_filter);

                attrsFromJob["search"] = this.model.searchJob.getSearch();
                attrsFromJob["sid"] = this.model.searchJob.entry.content.get('sid');
                attrsFromJob["dispatch.earliest_time"] = this.model.searchJob.getDispatchEarliestTimeOrAllTime();
                attrsFromJob["dispatch.latest_time"] = this.model.searchJob.getDispatchLatestTimeOrAllTime();
                this.mediateReportAttrsToClassicUrl(attrsFromJob);

                this.classicUrl.save(attrsFromJob, {
                    replaceState: true
                });
            },
            deleteOneTimeUseAttrs: function(attrs) {
                delete attrs["display.prefs.events.offset"];
                delete attrs["display.prefs.statistics.offset"];
                delete attrs["display.events.timelineEarliestTime"];
                delete attrs["display.events.timelineLatestTime"];
                delete attrs['display.statistics.sortColumn'];
                delete attrs['display.statistics.sortDirection'];
                delete attrs['display.events.table.sortColumn'];
                delete attrs['display.events.table.sortDirection'];
            },
            ensureTimeRangeFromReportInAttrs: function(attrs) {
                if ((attrs['dispatch.earliest_time'] !== void(0)) && (attrs['dispatch.latest_time'] === void(0))) {
                    attrs['dispatch.latest_time'] = this.model.report.entry.content.get("dispatch.latest_time");
                } else if ((attrs['dispatch.earliest_time'] === void(0)) && (attrs['dispatch.latest_time'] !== void(0))) {
                    attrs['dispatch.earliest_time'] = this.model.report.entry.content.get("dispatch.earliest_time");
                }
            },
            mediateReportAttrsToClassicUrl: function(attrs) {
                this.ensureTimeRangeFromReportInAttrs(attrs);
                
                general_utils.transferKey(attrs, 'search', 'q');
                general_utils.transferKey(attrs, 'dispatch.earliest_time', 'earliest');
                general_utils.transferKey(attrs, 'dispatch.latest_time', 'latest');
                
                if (attrs.q) {
                    attrs.q = splunkUtil.addLeadingSearchCommand(attrs.q);
                }
            },
            mediateClassicUrlAttrsToReport: function(attrs) {
                general_utils.transferKey(attrs, 'q', 'search');
                general_utils.transferKey(attrs, 'earliest', 'dispatch.earliest_time');
                general_utils.transferKey(attrs, 'latest', 'dispatch.latest_time');
                
                if ((attrs['dispatch.earliest_time'] !== void(0)) && (attrs['dispatch.latest_time'] === void(0))) {
                    if (time_utils.isRealtime(attrs['dispatch.earliest_time'])) {
                        attrs['dispatch.latest_time'] = "rt";
                    } else {
                        attrs['dispatch.latest_time'] = "";
                    }
                } else if ((attrs['dispatch.earliest_time'] === void(0)) && (attrs['dispatch.latest_time'] !== void(0))) {
                    if (time_utils.isRealtime(attrs['dispatch.latest_time'])) {
                        attrs['dispatch.earliest_time'] = "rt";
                    } else {
                        attrs['dispatch.earliest_time'] = "";
                    }
                }
                                
                if (attrs.search) {
                    attrs.search = splunkUtil.stripLeadingSearchCommand(attrs.search);
                }
            },
            applyModelFullPathToName: function(model, name) {
                //check to see if name is NOT the path to the entity
                if (!/^\/services.*/i.test(name)){
                    //we need to build the path to the entity from the SDK
                    name = splunkd_utils.fullpath(
                        model.url + "/" + encodeURIComponent(name),
                        {
                            app: this.model.application.get('app'),
                            owner: this.model.application.get('owner')
                        }
                    );
                }
                return name;
            },
            remove: function() {
                BaseController.prototype.remove.apply(this, arguments);
                
                var modelsToOmit = ['appLocal', 'user'],
                    collectionsToOmit = ['times', 'roles', 'eventRenderers', 'workflowActions'];
                
                if (!this.instantiatedReport) {
                    modelsToOmit.push("report");
                }
                
                if (!this.instantiatedTimeRange) {
                    modelsToOmit.push("timeRange");
                }

                if (!this.model.metaDataJob.isNew()) {
                    this.model.metaDataJob.destroy();
                }

                if (!this.model.searchJob.isNew()) {
                    if (this.model.searchJob.canBePausedOnRemove()) {
                        this.model.searchJob.pause();
                    }
                    this.model.searchJob.stopPolling();
                    this.model.searchJob.processKeepAlive = false;
                    this.model.searchJob.stopKeepAlive();
                }

                _.chain(this.model).omit(modelsToOmit).each(function(model) {
                    if (_.isFunction(model.deepOff)) {
                        model.deepOff();
                    }
                });
                
                _.chain(this.collection).omit(collectionsToOmit).each(function(collection) { 
                    if (_.isFunction(collection.deepOff)) {
                        collection.deepOff();
                    }
                });
            },
            close: function() {
                this.classicUrl.clear();
                this.classicUrl.save({}, {trigger: true});
            }
        });
    }
);
