define([
        'underscore', 
        'backbone', 
        'models/shared/Application',
        'models/services/AppLocal',
        'models/config',
        'models/services/server/ServerInfo',
        'models/services/data/ui/Nav',
        'models/services/authentication/User',
        'collections/services/AppLocals',
        'views/shared/Page',
        'util/csrf_protection',
        'util/ajax_no_cache',
        'util/ajax_logging',
        'util/splunkd_utils',
        'splunk.util',
        'util/console',
        'splunk.error'
    ], 
    function(
        _, 
        Backbone,
        ApplicationModel,
        AppLocalModel,
        configModel,
        ServerInfoModel,
        AppNavModel,
        UserModel,
        AppLocalsCollection,
        PageView,
        undefined,
        ajaxNoCache,
        ajaxLogging,
        splunkd_utils,
        splunkUtils,
        console,
        splunkError
    ) {
    return Backbone.Router.extend({
        routes: {
            ':locale/app/:app/:page': 'page',
            ':locale/app/:app/:page/': 'page',
            ':locale/app/:app/:page/*splat': 'page',
            ':locale/manager/:app/:page': 'page',
            ':locale/manager/:app/:page/': 'page',
            ':locale/manager/:app/:page/*splat': 'page',
            '*root/:locale/app/:app/:page': 'pageRooted',
            '*root/:locale/app/:app/:page/': 'pageRooted',
            '*root/:locale/app/:app/:page/*splat': 'pageRooted',
            '*root/:locale/manager/:app/:page': 'pageRooted',
            '*root/:locale/manager/:app/:page/': 'pageRooted',
            '*root/:locale/manager/:app/:page/*splat': 'pageRooted',
            '*splat': 'notFound' 
        },
        initialize: function(options) {
            //add __splunk__ to global namespace
            if (window.location.href.indexOf('debug=1')!=-1) {
                (function(exports) {
                    this.__splunk__ = exports;
                })(this);
            }
            //configuration
            this.enableAppBar = true;
            this.enablePageView = true;
            this.enableFooter = true;
            this.fetchUser = false;
            this.fetchAppLocals = false;
            this.fetchAppLocal = true;
            this.fetchServerInfo = true;
            
            this.loadingMessage = '';
            
            //models
            this.model = {};
            this.model.config = configModel;
            this.model.application = new ApplicationModel({owner: this.model.config.get('USERNAME')});
            this.model.user = new UserModel();
            this.model.appNav = new AppNavModel();
            this.model.appLocal = new AppLocalModel();
            this.model.serverInfo = new ServerInfoModel();
            
            //collections
            this.collection = {};
            this.collection.appLocals = new AppLocalsCollection();
            
            //views
            this.views = {};
            
            //deferreds
            this.deferreds = {};
            this.deferreds.user = $.Deferred();
            this.deferreds.appNav = $.Deferred();
            this.deferreds.appLocal = $.Deferred();
            this.deferreds.appLocals = $.Deferred();
            this.deferreds.serverInfo = $.Deferred();
            this.deferreds.pageViewRendered = $.Deferred();
            
            //history
            this.history = {};
            _.each(this.routes, function(value) {
                this.on('route:' + value, function() {
                    this.history[window.location.pathname] = true;
                }, this);
            }, this);
        },
        page: function(locale, app, page) {
            this.shouldRender = !this.history[window.location.pathname];
            this.model.application.set({
                locale: locale,
                app: app,
                page: page.split('?')[0]
            });
            
            this.bootstrapAppNav();
            this.bootstrapAppLocal();
            this.bootstrapServerInfo();
            
            if (this.enablePageView && !this.pageView) {
                this.bootstrapUser();
                this.bootstrapAppLocals();
                
                $.when(this.deferreds.user, this.deferreds.appLocals, this.deferreds.appLocal, this.deferreds.appNav, this.deferreds.serverInfo).then(function(){
                    this.pageView = new PageView({
                        splunkBar: true,
                        footer: this.enableFooter,
                        section: this.model.application.get('page'),
                        loadingMessage: this.loadingMessage,
                        model: {
                            application: this.model.application,
                            appNav: this.model.appNav,
                            appLocal: this.model.appLocal,
                            user: this.model.user,
                            serverInfo: this.model.serverInfo,
                            config: this.model.config
                        },
                        collection: {
                            apps: this.collection.appLocals
                        }
                    });
                    this.pageView.render();
                    this.deferreds.pageViewRendered.resolve();
                }.bind(this));
            }
        },
        pageRooted: function(root, locale, app, page) {
            this.model.application.set({
                root: root
            }, {silent: true});
            this.page(locale, app, page);
        },
        notFound: function() {
            console.log('Page not found.');
        },
        bootstrapAppNav: function() {
            var appNavPartialData;
            if (this.deferreds.appNav.state() !== 'resolved') {
                if (this.enableAppBar) {
                    appNavPartialData = __splunkd_partials__['/servicesNS/' + encodeURIComponent(this.model.application.get('owner')) + '/' + encodeURIComponent(this.model.application.get('app')) + '/apps/nav'];
                    if (appNavPartialData) {
                        this.model.appNav.setFromSplunkD(appNavPartialData);
                        this.deferreds.appNav.resolve();                        
                    } else {
                        this.model.appNav.fetch({
                            data: {
                                app: this.model.application.get("app"),
                                owner: this.model.application.get("owner")
                            },
                            success: function(model, response) {
                                this.deferreds.appNav.resolve();
                            }.bind(this),
                            error: function(model, response) {
                                this.deferreds.appNav.resolve();
                            }.bind(this)
                        });                        
                    }
                } else {
                    this.deferreds.appNav.resolve();
                    this.model.appNav = undefined;
                }
            }            
        },
        bootstrapAppLocals: function() {
            if (this.deferreds.appLocals.state() !== 'resolved') {
                if (this.fetchAppLocals) {
                    this.collection.appLocals.fetch({
                        data: {
                            sort_key: 'name',
                            sort_dir: 'asc',
                            app: '-' ,
                            owner: this.model.application.get('owner'),
                            search: 'visible=true AND disabled=0 AND name!=launcher',
                            count: -1
                        },
                        success: function(collection, response) {
                            this.deferreds.appLocals.resolve();
                        }.bind(this),
                        error: function(collection, response) {
                            this.deferreds.appLocals.resolve();
                        }.bind(this)
                    });
                } else {
                    this.deferreds.appLocals.resolve();
                    this.collection.appLocals = undefined;                    
                }
            }
        },
        bootstrapAppLocal: function() {
            var app = this.model.application.get('app'),
                appLocalPartialData;
            if (this.deferreds.appLocal.state() !== 'resolved') {
                if (this.fetchAppLocal && (app !== 'system')) {
                    appLocalPartialData = __splunkd_partials__['/servicesNS/nobody/system/apps/local/' + encodeURIComponent(app)];
                    if (appLocalPartialData) {
                        this.model.appLocal.setFromSplunkD(appLocalPartialData);
                        this.deferreds.appLocal.resolve();
                    } else {
                        this.model.appLocal.fetch({
                            url: splunkd_utils.fullpath(this.model.appLocal.url + "/" + encodeURIComponent(app)),
                            data: {
                                app: app,
                                owner: this.model.application.get("owner")
                            },
                            success: function(model, response) {
                                this.deferreds.appLocal.resolve();
                            }.bind(this),
                            error: function(model, response) {
                                this.deferreds.appLocal.resolve();
                            }.bind(this)
                        });
                    }                    
                } else {
                    this.deferreds.appLocal.resolve();
                    this.model.appLocal = undefined;
                }
            }            
        },
        bootstrapServerInfo: function() {
            var serverInfoPartialData;
            if (this.deferreds.serverInfo.state() !== 'resolved') {
                if (this.fetchServerInfo) {
                    serverInfoPartialData = __splunkd_partials__['/services/server/info'];
                    if (serverInfoPartialData) {
                        this.model.serverInfo.setFromSplunkD(serverInfoPartialData);
                        this.deferreds.serverInfo.resolve();
                    } else {
                        this.model.serverInfo.fetch({
                            success: function(model, response) {
                                this.deferreds.serverInfo.resolve();
                            }.bind(this),
                            error: function(model, response) {
                                this.deferreds.serverInfo.resolve();
                            }.bind(this)                        
                        });
                    }
                } else {
                    this.deferreds.serverInfo.resolve();
                    this.model.serverInfo = undefined;                    
                }
            }  
        },
        bootstrapUser: function() {
            if (this.deferreds.user.state() !== 'resolved') {
                if (this.fetchUser) {
                    this.model.user.fetch({
                        url: splunkd_utils.fullpath(this.model.user.url + "/" + encodeURIComponent(this.model.application.get("owner"))),
                        data: {
                            app: this.model.application.get("app"),
                            owner: this.model.application.get("owner")
                        },
                        success: function(model, response) {
                            this.deferreds.user.resolve();
                        }.bind(this),
                        error: function(model, response) {
                            this.deferreds.user.resolve();
                        }.bind(this)
                    });
                } else {
                    this.deferreds.user.resolve();
                    this.model.user = undefined;
                }
            }
        },
        // convenience method for subclasses to update the page <title>
        // as browser compatibility issues arise, they can be encapsulated here
        setPageTitle: function(title) {
            this.deferreds.serverInfo.done(function(){
                var version = this.model.serverInfo.entry.content.get('version') || _('N/A').t();
                document.title = splunkUtils.sprintf(_('%s | Splunk %s').t(), title, version);
            }.bind(this));
        }
    });
});
