define(function(require) {
    var _ = require('underscore');
    var $ = require('jquery');
    var Backbone = require('backbone');
    var FieldsetView = require('../../simpleform/fieldsetview');
    /**
     * Delegate view for dashboard panels that deals with hiding/showing the drag-handle for edit mode as well as
     * hiding the panel if all dashboard elements it contains are hidden.
     */
    var DashboardPanelView = Backbone.View.extend({
        initialize: function() {
            this.listenTo(this.model, 'change:edit', this.onEditStateChange, this);
        },
        onElementVisibilityChange: function() {
            if (this.$('.dashboard-element:not(.hidden)').length === 0 && (this.fieldset == null || this.fieldset.isEmpty())) {
                // Hide the panel if all dashboard elements are hidden
                this.hide();
            } else {
                this.show();
            }
            var isEditMode = this.model.get('edit');
            _(this.$el.children('.panel-element-row.grouped')).each(function(panelRow) {
                var selector = '.dashboard-element' + (isEditMode ? '' : ':not(.hidden)');
                var elements = $(panelRow).children(selector);
                if (elements.length > 0) {
                    elements.css({ width: String(100 / elements.length) + '%' });
                }
            });
        },
        hide: function() {
            if (!this.$el.is('.hidden')) {
                if (!this.model.get('edit')) {
                    this.$el.hide();
                }
                this.$el.addClass('hidden').trigger('panelVisibilityChanged');
            }
        },
        show: function() {
            if (this.$el.is('.hidden')) {
                this.$el.show().removeClass('hidden').trigger('panelVisibilityChanged');
            }
        },
        onEditStateChange: function(model) {
            if (model.get('edit')) {
                if (!this._dragHandle) {
                    this._dragHandle = $('<div class="drag-handle"><div class="handle-inner"></div></div>');
                }
                this._dragHandle.prependTo(this.el);
                if (this.$el.is('.hidden')) {
                    this.$el.show().trigger('panelVisibilityChanged');
                }
                if (!this.fieldset) {
                    this.fieldset = new FieldsetView({
                        el: $('<div class="fieldset"></div>').insertAfter(this._dragHandle)
                    }).render();
                }
            } else {
                if (this._dragHandle) {
                    this._dragHandle.detach();
                }
                if (this.$el.is('.hidden')) {
                    this.$el.hide().trigger('panelVisibilityChanged');
                }
                if (this.fieldset && this.fieldset.isEmpty()) {
                    this.fieldset.remove();
                    this.fieldset = null;
                }
            }
            this.onElementVisibilityChange();
        },
        markGroupedElementRows: function(){
            // Find all panel-rows with more than 1 element in order to re-layout them on visibility changes
            _(this.$el.children('.panel-element-row'))
                .chain()
                .map($)
                .invoke('removeClass', 'grouped')
                .filter(function($el){ return $el.children('.dashboard-element').length > 1; })
                .invoke('addClass', 'grouped');
        },
        render: function() {
            if ((!this.fieldset)) {
                var fieldsetEl = this.$el.children('.fieldset');
                if (fieldsetEl.length) {
                    this.fieldset = new FieldsetView({ el: fieldsetEl }).render();
                }
            }
            this.onEditStateChange(this.model);
            this.$el.off('elementVisibilityChanged');
            this.markGroupedElementRows();
            this.$el.on('elementVisibilityChanged', _.bind(this.onElementVisibilityChange, this));
            return this;
        },
        isEmpty: function() {
            var panelElementRows = this.$el.children('.panel-element-row');
            var inputs = this.$el.children('.fieldset').children('.input');
            return panelElementRows.length === 0 && inputs.length === 0;
        },
        events: {
            'itemRemoved': function(e) {
                if (this.isEmpty()) {
                    this.remove();
                }
            }
        },
        remove: function() {
            if (this.fieldset) {
                this.fieldset.remove();
            }
            this.$el.off('elementVisibilityChanged');
            var parent = this.$el.parent();
            Backbone.View.prototype.remove.call(this);
            parent.trigger('panelRemoved');
        }
    });

    return DashboardPanelView;
});