define(function(require) {
    var _ = require('underscore'), $ = require('jquery');
    var mvc = require('../../mvc');
    var BaseDashboardModel = require('models/search/Dashboard');
    var xmlUtils = require('util/xml');
    var GeneralUtils = require('util/general_utils');
    var console = require('util/console');

    var DashboardModel = BaseDashboardModel.extend({
        initialize: function() {
            BaseDashboardModel.prototype.initialize.apply(this, arguments);
        },
        /**
         *
         * @param options {
         *     tokens (boolean) -  whether the generated XML source should contain tokens or their values
         *     useLoadjob (boolean) - whether to use | loadjob <sid> instead of writing the actual search to the XML
         *     indent (boolean) - whether to generated pretty-printed XML or not
         * }
         * @returns {String} the serialized XML
         */
        getFlattenedXML: function(options) {
            if (!this._structure) {
                throw new Error('Cannot create flattened XML without the item order being captured first');
            }
            var $n = xmlUtils.$node, that = this;
            options = options || {};

            var dashboard = $n('<dashboard/>');
            $n('<label/>').text(this.meta.get('label')).appendTo(dashboard);
            if (this.meta.has('description')) {
                $n('<description/>').text(this.meta.get('description')).appendTo(dashboard);
            }

            //TODO add serialization of form elements (condition for pdf printing?)
            _(this._structure.rows).each(function(panels) {
                var row = $n('<row/>');

                if (_(panels).any(function(p) { return p.elements.length > 1; })) { // Are there any grouped panels?
                    var groups = _(panels).map(function(p) { return p.elements.length; }).join(',');
                    row.attr('grouping', groups);
                }

                _(panels).chain().flatten().pluck('elements').flatten().each(function(id) {
                    var element = mvc.Components.get(id);
                    var settings = element.model.mapToXML(_.extend({ tokens: false, flatten: true }, options));
                    var newNode = xmlUtils.$node('<' + settings.type + '/>');

                    if (settings.attributes !== undefined) {
                        newNode.attr(settings.attributes);
                    }

                    if (settings.content) {
                        if (settings.cdata) {
                            newNode.append(xmlUtils.cdata(settings.content));
                        } else {
                            newNode.text(settings.content);
                        }
                    } else {
                        that._applyTitle(newNode, settings);
                        var manager = mvc.Components.get(element.model.entry.content.get('display.general.manager'));
                        if (options.useLoadjob !== false && manager && manager.job) {
                            that._applySearch(newNode, {
                                search: {
                                    type: 'inline',
                                    search: '| loadjob ' + manager.job.sid + ' ignore_running=t'
                                }
                            });
                        } else {
                            that._applySearch(newNode, settings);
                        }

                        that._applyOptions(newNode, settings);
                        that._applyTags(newNode, settings);
                    }

                    newNode.appendTo(row);
                });

                row.appendTo(dashboard);

            });

            if (options && options.indent) {
                this.formatXML(dashboard);
            }

            return xmlUtils.serialize(dashboard);
        },
        /**
         * Update the dashboard element with the settings specified
         * @param id - the element id (string)
         * @param settings - a settings object containing the update information
         * @param options - persistence options
         *
         * settings contains: {
         *      type: (String) the element type (one of "table", "chart", "single", "map", "list", "html")
         *      search: (Object) {
         *          type: (String) one of 'inline', 'saved' or 'pivot'
         *          search: (String) the search string (for inline)
         *          earliest_time: (String) the earliest_time of the inline search
         *          latest_time: (String) the latest_time of the inline search
         *          name: (String) name of the saved search
         *          ... TODO pivot search params
         *      }
         *      options: (Object) options to added (or replaced) to the element (<option name="$name$">$value$</option>)
         *      removeOptions: (Array) options to be removed from the xml element
         *      tags: (Object) tags to be added (or replaced) to the element (<$tag$>$value$</$tag$>)
         * },
         * options: {
         *      clearOptions: (Boolean) true to remove all options nodes before updating the XML element
         * }
         */
        updateElement: function(id, settings, options) {
            console.log('About to update item=%o with settings=%o', id, settings);
            var $xml = this.get$XML();
            var cur = this.getXMLNode(id, $xml);
            if (!cur) {
                throw new Error("Unable to find dashboard element with ID " + id);
            }
            var newNode = xmlUtils.$node('<' + settings.type + '/>');
            // Transfer allowed attributes to the new XML node
            _(['id', 'depends', 'rejects']).each(function(attr) {
                var val = $(cur).attr(attr);
                if (val) {
                    newNode.attr(attr, val);
                }
            });
            newNode.append($(cur).children());

            if (options && options.clearOptions) {
                newNode.find('option').remove();
                delete settings.options;
            }

            this._applyTitle(newNode, settings);
            if (settings.search && settings.search.type === 'global') {
                this._applyGlobalSearch($xml, settings);
            } else {
                this._applySearch(newNode, settings);
            }
            this._applyOptions(newNode, settings);
            this._applyTags(newNode, settings);

            $(cur).replaceWith(newNode);
            if (console.DEBUG_ENABLED) {
                this.formatXML($xml);
                console.log(xmlUtils.serialize($xml));
            }
            this.set$XML($xml);
            return this.save();
        },
        deleteElement: function(id) {
            // Create new item order without the element which is to be deleted
            var newStructure = {
                fieldset: this._structure.fieldset,
                rows: _(this._structure.rows).map(function(row) {
                    return _(row).map(function(panel) {
                        return {
                            inputs: panel.inputs,
                            elements: _(panel.elements).without(id)
                        };
                    });
                })
            };
            return this.updateStructure(newStructure);
        },
        addElement: function(id, settings) {
            var row = xmlUtils.$node('<row/>');
            var panel = xmlUtils.$node('<panel/>').appendTo(row);
            var newNode = xmlUtils.$node('<' + settings.type + '/>').appendTo(panel);
            this._structure.rows.push([
                {
                    inputs: [],
                    elements: [settings.id]
                }
            ]);
            this._applyTitle(newNode, settings);
            this._applySearch(newNode, settings);
            var $xml = this.get$XML();
            $xml.find(':eq(0)').append(row);
            if (console.DEBUG_ENABLED) {
                this.formatXML($xml);
                console.log(xmlUtils.serialize($xml));
            }
            this.set$XML($xml);
            return this.save();
        },
        _migrateViewType: function($xml, tagName) {
            var curRoot = $xml.find(':eq(0)');
            if (curRoot.prop('tagName') !== tagName) {
                $xml = $($.parseXML('<' + tagName + '/>'));
                var newRoot = $xml.find(':eq(0)'), cur = curRoot[0];
                _(cur.attributes).each(function(attr) {
                    newRoot.attr(attr.nodeName, attr.nodeValue);
                });
                while (cur.childNodes.length) {
                    newRoot.append(cur.childNodes[0]);
                }
                this.trigger('change:rootNodeName', this, tagName);
            }
            return $xml;
        },
        _applyOptions: function(newNode, settings) {
            if (settings.options) {
                _.each(settings.options, function(value, name) {
                    var curOption = newNode.find('option[name="' + name + '"]');
                    if (value === "" || value === null || value === void(0)) {
                        curOption.remove();
                    } else {
                        if (curOption.length) {
                            curOption.text(value);
                        } else {
                            xmlUtils.$node('<option/>').attr('name', name).text(value).appendTo(newNode);
                        }
                    }
                });
            }
            if (settings.removeOptions) {
                _(settings.removeOptions).each(function(name) {
                    newNode.find('option[name="' + name + '"]').remove();
                });
            }
        },
        _applyTags: function(newNode, settings) {
            if (settings.tags) {
                _.each(settings.tags, function(value, tag) {
                    newNode.find(tag).remove();
                    if ((_.isArray(value) && value.length) || value) {
                        xmlUtils.$node('<' + tag + '/>').text(value).appendTo(newNode);
                    }
                });
            }
        },
        _applyTitle: function(newNode, settings) {
            var titleNode = newNode.find('title');
            if (settings.title) {
                if (!titleNode.length) {
                    titleNode = xmlUtils.$node('<title/>').prependTo(newNode);
                }
                titleNode.text(settings.title);
            } else {
                titleNode.remove();
            }
        },
        _applySearch: function(newNode, settings) {
            if (settings.search) {
                // Clear current search info
                newNode.find('searchString,searchTemplate,searchName,searchPostProcess,pivotSearch,earliestTime,latestTime').remove();
                var titleNode = newNode.find('title');
                switch (settings.search.type) {
                    case 'inline':
                    case 'global': // Apply global search as an inline search for flattened XML
                        var searchNode = xmlUtils.$node('<searchString/>').text(settings.search.search);
                        if (titleNode.length) {
                            searchNode.insertAfter(titleNode);
                        } else {
                            searchNode.prependTo(newNode);
                        }
                        if (settings.search.latest_time !== undefined) {
                            xmlUtils.$node('<latestTime/>').text(settings.search.latest_time).insertAfter(searchNode);
                        }
                        if (settings.search.earliest_time !== undefined) {
                            xmlUtils.$node('<earliestTime/>').text(settings.search.earliest_time).insertAfter(searchNode);
                        }
                        break;
                    case 'postprocess':
                        var postSearchNode = xmlUtils.$node('<searchPostProcess/>').text(settings.search.search);
                        if (titleNode.length) {
                            postSearchNode.insertAfter(titleNode);
                        } else {
                            postSearchNode.prependTo(newNode);
                        }
                        break;
                    case 'saved':
                        var savedNode = xmlUtils.$node('<searchName/>').text(settings.search.name);
                        if (titleNode.length) {
                            savedNode.insertAfter(titleNode);
                        } else {
                            savedNode.prependTo(newNode);
                        }
                        break;
                    case 'pivot':
                        throw new Error("Pivot search not implemented!");
                    default:
                        throw new Error("Unknown search type: " + settings.search.type);
                }
            }
        },
        _applyGlobalSearch: function($xml, settings) {
            var root = $xml.find(':eq(0)');
            root.children('searchTemplate,earliestTime,latestTime').remove();
            var labelNode = root.children('label');
            var descNode = root.children('description');

            var searchNode = xmlUtils.$node('<searchTemplate/>').text(settings.search.search);
            if (descNode.length) {
                searchNode.insertAfter(descNode);
            } else if (labelNode.length) {
                searchNode.insertAfter(labelNode);
            } else {
                searchNode.prependTo(root);
            }
            if (settings.search.latest_time) {
                xmlUtils.$node('<latestTime/>').text(settings.search.latest_time).insertAfter(searchNode);
            }
            if (settings.search.earliest_time) {
                xmlUtils.$node('<earliestTime/>').text(settings.search.earliest_time).insertAfter(searchNode);
            }
        },
        captureDashboardStructure: function(structure) {
            var $xml = this.get$XML();
            this.validateItemOrder(this._getItemOrderMap(structure.rows, $xml));
            this.validateInputOrder(this._getInputOrderMap(structure, $xml));
            this._structure = structure;
        },
        getStructureMap: function(structure, $xml) {
            return _.extend({},
                this._getItemOrderMap(structure.rows, $xml),
                this._getInputOrderMap(structure, $xml)
            );
        },
        getXMLNode: function(id, $xml) {
            id = String(id);
            var node = $xml.find(':eq(0)>fieldset>#' + id + ',:eq(0)>row>#' + id + ',:eq(0)>row>panel>#' + id);
            if (node.length) {
                return node[0];
            } else {
                return this.getStructureMap(this._structure, $xml)[id];
            }
        },
        _getItemOrderMap: function(rowOrder, $xml) {
            var elements = _($xml.find('row').children()).map(function(el) {
                return $(el).is('panel') ? $(el).children(':not(input)').toArray() : el;
            });
            var makeObject = function(keys, values, hint) {
                if (keys.length !== values.length) {
                    throw new Error('Item order did not match XML structure (' + keys.length + ' keys, ' + values.length + ' nodes)');
                }
                return _.object(keys, values);
            };
            return makeObject(_(rowOrder).chain().flatten().pluck('elements').flatten().value(), _.flatten(elements));
        },
        validateItemOrder: function(itemOrderMap) {
            _(itemOrderMap).each(function(node, id) {
                var nid = $(node).attr(id);
                if (!(nid === undefined || nid === id)) {
                    throw new Error('Invalid Item order. Expected node with ID ' + id + '. Instead saw ' + nid);
                }
            });
        },
        inputSettingsToXML: function(settingsModel, inputNode) {
            var $input = inputNode ? $(inputNode) : xmlUtils.$node('<input/>');
            var settings = settingsModel.toJSON({tokens: true});
            $input.attr('type', settings.type);
            $input.attr('token', settings.token || null);
            $input.attr('searchWhenChanged', settings.searchWhenChanged != null ?
                String(GeneralUtils.normalizeBoolean(settings.searchWhenChanged)) : null);
            // set the label if it is different than the token name
            var label = $.trim(settings.label);
            if (label !== undefined && label !== settings.token) {
                xmlUtils.replaceOrPrepend({ node: xmlUtils.$node('<label/>').text(label), selector: 'label', container: $input });
            } else {
                $input.children('label').remove();
            }
            // Add static choices
            if (settings.choices && settings.choices.length) {
                var newChoices = _(settings.choices).map(function(choice) {
                    return xmlUtils.$node('<choice/>').attr('value', choice.value).text(choice.label);
                });
                xmlUtils.replaceOrAppend({ node: newChoices, container: $input, selector: 'choice' });
            } else {
                $input.children('choice').remove();
            }
            // Add the search
            this._applyPopulatingSearch(settings, $input);
            // default is a named node
            var defaultValue = settings['default'];
            if (defaultValue) {
                var defaultNode = xmlUtils.$node('<default/>');
                if (settings.type == 'time') {
                    xmlUtils.$node('<earliestTime></earliestTime>').text((defaultValue.earliest_time == null) ? '' : defaultValue.earliest_time).appendTo(defaultNode);
                    xmlUtils.$node('<latestTime></latestTime>').text((defaultValue.latest_time  == null) ? '' : defaultValue.latest_time).appendTo(defaultNode);
                } else {
                    defaultNode.text(defaultValue);
                }
                xmlUtils.replaceOrAppend({ node: defaultNode, selector: 'default', container: $input });
            } else {
                $input.children('default').remove();
            }
            _(['prefix', 'suffix', 'seed', 'valuePrefix', 'valueSuffix', 'delimiter']).each(function(option) {
                var val = settings[option];
                if (val) {
                    var node = xmlUtils.$node('<' + option + '/>').text(val);
                    xmlUtils.replaceOrAppend({ node: node, container: $input, selector: option });
                } else {
                    $input.children(option).remove();
                }
            });
            return $input;
        },
        _applyPopulatingSearch: function(settings, $input) {
            var $search;
            if (settings.searchType === 'saved') {
                if (!settings.searchName) {
                    $input.children('populatingSavedSearch,populatingSearch').remove();
                    return;
                }
                $search = xmlUtils.$node('<populatingSavedSearch/>');
                $search.text(settings.searchName);
            } else {
                if (!settings.search) {
                    $input.children('populatingSavedSearch,populatingSearch').remove();
                    return;
                }
                $search = xmlUtils.$node('<populatingSearch/>');
                $search.text(settings.search);
                $search.attr('earliest', settings['earliest_time']);
                $search.attr('latest', settings['latest_time']);
            }
            if (settings.labelField) {
                $search.attr('fieldForLabel', settings.labelField);
            }
            if (settings.valueField) {
                $search.attr('fieldForValue', settings.valueField);
            }
            xmlUtils.replaceOrAppend({ node: $search, container: $input, selector: 'populatingSavedSearch,populatingSearch'});
        },
        updateInputXML: function(id, settings, options) {
            console.log('About to update input=%o with settings=%o', id, settings.toJSON());
            var $xml = this.get$XML();
            var cur = this.getXMLNode(id, $xml);
            if (!cur) {
                throw new Error('Unable to find input with ID=', id);
            }
            cur = $(cur);
            var newNode = this.inputSettingsToXML(settings, cur);
            $(cur).replaceWith(newNode);
            if (console.DEBUG_ENABLED) {
                this.formatXML($xml);
                console.log(xmlUtils.serialize($xml));
            }
            this.set$XML($xml);
            return this.save();
        },
        isFormAutoRun: function() {
            var $xml = this.get$XML();
            var fieldset = $xml.find('fieldset');
            return GeneralUtils.normalizeBoolean(fieldset.attr('autoRun'));
        },
        updateFormSettings: function(options) {
            var $xml = this._migrateViewType(this.get$XML(), 'form');
            var fieldset = $xml.find('fieldset');
            if (!fieldset.length) {
                fieldset = this._createNewFieldset($xml);
            }
            if (options.hasOwnProperty('submitButton')) {
                fieldset.attr('submitButton', String(options.submitButton));
            }
            if (options.hasOwnProperty('autoRun')) {
                fieldset.attr('autoRun', String(options.autoRun));
            }
            this.set$XML($xml);
            return this.save();
        },
        _createNewFieldset: function($xml) {
            var $fieldset = xmlUtils.$node('<fieldset submitButton="false"></fieldset>');
            xmlUtils.inject({
                node: $fieldset,
                where: 'after',
                container: $xml.find(':eq(0)'),
                selectors: ['description', 'label'],
                fallback: 'prepend'
            });
            return $fieldset;
        },
        deleteInput: function(id) {
            var newStructure = {
                fieldset: _(this._structure.fieldset).without(id),
                rows: _(this._structure.rows).map(function(row) {
                    return _(row).map(function(panel) {
                        return {
                            inputs: _(panel.inputs).without(id),
                            elements: panel.elements
                        };
                    });
                })
            };
            return this.updateStructure(newStructure);
        },
        _getInputOrderMap: function(structure, $xml) {
            var map = {};
            var makeObject = function(keys, values, hint) {
                if (keys.length !== values.length) {
                    throw new Error('Input (' + hint + ') order did not match XML structure (' + keys.length + ' keys, ' + values.length + ' nodes)');
                }
                return _.object(keys, values);
            };
            _.extend(map, makeObject(structure.fieldset, $xml.find(':eq(0)>fieldset>input,:eq(0)>fieldset>html'), 'fieldset'));
            _.extend(map, makeObject(_(structure.rows).chain().flatten().pluck('inputs').flatten().value(), $xml.find(':eq(0)>row>panel>input'), 'panels'));
            return map;
        },
        validateInputOrder: function(inputOrderMap) {
            _(inputOrderMap).each(function(node, id) {
                var nid = $(node).attr(id);
                if (!(nid === undefined || nid === id)) {
                    throw new Error('Invalid Input order. Expected node with ID ' + id + '. Instead saw ' + nid);
                }
            });
        },
        updateInput: function(settingsModel) {
            return this.updateInputXML(settingsModel.get('name'), settingsModel, {});
        },
        addInput: function(settingsModel) {
            var $input = this.inputSettingsToXML(settingsModel);
            // Ensure that we are a form
            var $xml = this._migrateViewType(this.get$XML(), 'form');
            // append xml to fieldset
            var $fieldset = $xml.find(':eq(0)>fieldset');
            if (!$fieldset.length) {
                $fieldset = this._createNewFieldset($xml);
            }
            $fieldset.append($input);
            // ensure the mapping and order is updated
            this._structure.fieldset.push(settingsModel.get('id'));
            this.set$XML($xml);
            return this.save();
        },
        updateStructure: function(structure) {
            if (!this._structure) {
                console.warn('No captured dashboard structure');
                this.captureDashboardStructure(structure);
            }

            if (!_.isEqual(structure, this._structure)) {
                // Remove empty panels from structure
                structure.rows = _(structure.rows).map(function(row) {
                    return _(row).filter(function(panel) {
                        return panel.elements.length > 0 || panel.inputs.length > 0;
                    });
                });
                // Remove empty rows from structure
                structure.rows = _(structure.rows).filter(function(row) {
                    return row.length > 0;
                });

                var $node = xmlUtils.$node;
                var $xml = this.get$XML();
                var rootNode = $xml.find(':eq(0)');

                // Update order of dashboard elements
                var itemMap = this._getItemOrderMap(this._structure.rows, $xml);
                var inputMap = this._getInputOrderMap(this._structure, $xml);

                rootNode.children('row').detach();
                _(structure.rows).each(function(rowOrder) {
                    var row = $node('<row/>');
                    _(rowOrder).each(function(panelOrder) {
                        var panel = $node('<panel/>');
                        _(panelOrder.elements).each(function(elementId) {
                            var item = itemMap[elementId];
                            $(item).find('script').remove();
                            panel.append(item);
                        });
                        row.append(panel);
                    });
                    rootNode.append(row);
                });

                // Update inputs in fieldset
                var fieldset = rootNode.children('fieldset');
                if (!fieldset.length) {
                    fieldset = this._createNewFieldset($xml);
                }
                fieldset.children().detach();
                _(structure.fieldset).each(function(inputId) {
                    var inputNode = inputMap[inputId];
                    if (inputNode) {
                        $(inputNode).find('script').remove();
                        fieldset.append(inputNode);
                    } else {
                        console.warn('input not found', inputId);
                    }
                });
                // remove the fieldset if it is empty
                if (fieldset.children().length == 0) {
                    fieldset.remove();
                }

                // Update inputs within panels
                var rows = rootNode.children('row');
                _(structure.rows).each(function(rowOrder, rowIndex) {
                    var row = rows[rowIndex];
                    var panels = $(row).children('panel');
                    _(rowOrder).each(function(panelOrder, panelIndex) {
                        var panel = panels[panelIndex];
                        // Prepend inputs in reverse order to the panel
                        _(panelOrder.inputs).chain().clone().reverse().each(function(inputId) {
                            var inputNode = inputMap[inputId];
                            $(inputNode).find('script').remove();
                            $(panel).prepend(inputNode);
                        });
                    });
                });

                console.log('Updated dashboard structure', JSON.stringify(structure), $xml.find(':eq(0)')[0]);
                this._structure = structure;

                // Remove empty panels and rows
                rootNode.children('row>panel:empty').remove();
                rootNode.children('row:empty').remove();
                if (this.hasInputs()) {
                    $xml = this._migrateViewType($xml, 'form');
                } else {
                    $xml = this._migrateViewType($xml, 'dashboard');
                }
                this.set$XML($xml);
                console.log('Saving XML changes');
                return this.save();
            } else {
                console.log('no changes to input order');
                return $.Deferred().resolve();
            }
        },
        isEditable: function() {
            return this.isDashboard() || this.isForm();
        },
        getRootNodeName: function() {
            return (this.get$XML().find(':eq(0)').prop('nodeName') || '').toLowerCase();
        },
        hasInputs: function(){
            return this._structure.fieldset.length > 0 || _(this._structure.rows).any(function(row){
                return _(row).any(function(element){
                    return element.inputs && element.inputs.length > 0;
                });
            });
        }
    });

    return DashboardModel;
});