define(function(require, module, exports) {
    var _ = require('underscore'), $ = require('jquery'), mvc = require('splunkjs/mvc');
    var Modal = require('views/shared/Modal');
    var BaseModel = require('models/Base');
    var utils = require('../../utils');
    var Dashboard = require('../controller');
    var DashboardElement = require('../element/base');
    var Mapper = require('../mapper');
    var AddPanelView = require('./addpanel/master');
    var SearchManager = require('../../searchmanager');
    var SavedSearchManager = require('../../savedsearchmanager');
    var console = require('util/console');
    var TimeRangePickerView = require('views/shared/timerangepicker/dialog/Master');
    var TimeRangeModel = require('models/shared/TimeRange');
    var TimeRangeDelegate = require('views/shared/delegates/ModalTimerangePicker');
    var controller = require('../controller');
    var route = require('uri/route');
    var FlashMessages = require('views/shared/FlashMessages'); 
    
    /**
     * Transient model representing the information for a new dashboard panel element
     */
    var NewPanelModel = BaseModel.extend({
        defaults: {
            elementCreateType: 'inline',
            'dispatch.earliest_time': '0',
            'dispatch.latest_time': ''
        },
        validation: {
            search: {
                fn: 'validateSearchQuery'
            }
        },
        validateSearchQuery: function(value, attr, computedState) {
            if(computedState['elementCreateType'] === 'inline' && !value) {
                return 'Search string is required.';
            }
        },
        sync: function(method, model, options) {
            console.log('NewPanelModel.sync(%o, %o, %o)', method, model, options);
            if(method !== 'create') {
                throw new Error('Unsupported sync method: ' + method);
            }
            if(!model.isValid()) {
                return false;
            }
            var dfd = $.Deferred();
            var searchType = this.get('elementCreateType');
            var elementId, i = 1;
            do {
                elementId = 'element' + (i++);
            } while(mvc.Components.has(elementId));

            var vizType = searchType === 'saved' ? this.get('savedSearchVisualization') || 'visualizations:charting' : 'visualizations:charting',
                    mapper = Mapper.get(vizType);

            Dashboard.getStateModel().view.addElement(elementId, {
                type: mapper.tagName,
                title: this.get('title'),
                search: {
                    type: searchType,
                    search: this.get('search'),
                    earliest_time: this.get('dispatch.earliest_time'),
                    latest_time: this.get('dispatch.latest_time'),
                    name: this.get('savedSearchName')
                }
            }).done(_.bind(function() {
                        var newItemElement = mvc.Components.get('dashboard').createNewElement({
                            title: this.get('title'),
                            id: elementId
                        });
                        var newSearchId = _.uniqueId('new-search');
                        switch(searchType) {
                            case 'inline':
                                new SearchManager({
                                    "id": newSearchId,
                                    "search": this.get('search'),
                                    "earliest_time": this.get('dispatch.earliest_time') || "0",
                                    "latest_time": this.get('dispatch.latest_time') || '',
                                    "app": utils.getCurrentApp(),
                                    "auto_cancel": 90,
                                    "status_buckets": 0,
                                    "preview": true,
                                    "timeFormat": "%s.%Q",
                                    "wait": 0,
                                    "runOnSubmit": true
                                }, {tokens: true});
                                break;
                            case 'saved':
                                new SavedSearchManager({
                                    "id": newSearchId,
                                    "searchname": this.get('savedSearchName'),
                                    "app": utils.getCurrentApp(),
                                    "auto_cancel": 90,
                                    "status_buckets": 0,
                                    "preview": true,
                                    "timeFormat": "%s.%Q",
                                    "wait": 0
                                });
                                break;
                        }

                        var ElementType = DashboardElement.extend({
                            initialVisualization: vizType
                        });

                        var component = new ElementType({
                            id: elementId,
                            managerid: newSearchId,
                            el: newItemElement
                        }, {tokens: true});
                        component.render();
                        dfd.resolve();
                    }, this));

            return dfd.promise();
        }
    });
    
    var PanelTimeRangeModel = TimeRangeModel.extend({
        validation: _.extend({
            earliest_token: function(value, attr, computedState) {
                if(computedState.useTimeFrom === 'tokens' && !value) {
                    return 'No value specified for earliest token.';
                }
            },
            latest_token: function(value, attr, computedState) {
                if(computedState.useTimeFrom === 'tokens' && !value) {
                    return 'No value specified for latest token.';
                }
            }
        })
    });

    return Modal.extend({
        moduleId: module.id,
        className: 'modal add-panel',
        initialize: function() {
            Modal.prototype.initialize.apply(this, arguments);
            this.model=  this.model || {};
            this.model.report = new NewPanelModel();
            this.model.timeRange = new PanelTimeRangeModel({
                'earliest': "0",
                'latest': ""
            });

            var appModel = Dashboard.model.app;
            var userModel = Dashboard.model.user;
            var appLocalModel = Dashboard.model.appLocal;
            var timesCollection = Dashboard.collection.times;

            this.children.addPanel = new AddPanelView({ 
                model: {
                    report: this.model.report,
                    timeRange: this.model.timeRange,
                    state: controller.getStateModel()
                }, 
                collection: {
                    times: timesCollection
                }, 
                controller: this.options.controller  
            });

            this.children.timeRangePickerView = new TimeRangePickerView({
                model: {
                    timeRange: this.model.timeRange,
                    user: userModel,
                    appLocal: appLocalModel,
                    application: appModel
                },
                collection: timesCollection,
                appendSelectDropdownsTo: '.modal:visible'
            });

            this.children.flashMessages = new FlashMessages({model: this.model});

            this.model.timeRange.on('applied', function() {
                this.timeRangeDelegate.closeTimeRangePicker();
            }, this);

            this.listenTo(this.model.report, 'change:elementCreateType', this.handleSubmitButtonState, this);

        },
        events: {
            'click a.modal-btn-primary': function(e) {
                if(this.model.report.get('elementCreateType') === 'pivot') {
                    return;
                }
                e.preventDefault();
                Dashboard.getStateModel().set('edit', false);
                var modal = this;
                var useTimeFrom = this.model.timeRange.get('useTimeFrom');
                var timeTokenPrefix = useTimeFrom == "global" ? '': useTimeFrom + '.';
                var errors = this.model.report.validate();

                if (useTimeFrom === "tokens") {
                    var timeRangeErrors = this.model.timeRange.validate();
                    if (timeRangeErrors) {
                        errors = (errors) ? _.extend(errors, timeRangeErrors) : timeRangeErrors;
                    }
                }

                if (errors) {
                    this.children.flashMessages.flashMsgCollection.reset();
                    _.each(errors, function(val) {
                        this.children.flashMessages.flashMsgCollection.add({
                            key: 'addpanel-' + val,
                            type: 'error',
                            html: _.escape(_(val).t())
                        });
                    }, this);
                    return;
                }

                if (useTimeFrom === "tokens") {
                    this.model.report.set({
                        'dispatch.earliest_time': '$' + this.model.timeRange.get('earliest_token') +'$',
                        'dispatch.latest_time': '$' + this.model.timeRange.get('latest_token') +'$',
                        'useTimeFrom': 'tokens'
                    }, {tokens: true});
                } else if (useTimeFrom !== "search"){
                    this.model.report.set({
                        'dispatch.earliest_time': '$' + timeTokenPrefix +'earliest$',
                        'dispatch.latest_time': '$' + timeTokenPrefix +'latest$'
                    }, {tokens: true});
                }
                var dfd = this.model.report.save();

                if(dfd) {
                    dfd.done(function() {
                        setTimeout(function() {
                            Dashboard.getStateModel().set('edit', true);
                        }, 250);
                        modal.hide();
                    });
                } else {
                    Dashboard.getStateModel().set('edit', true);
                }
            }, 
            'hide': function(e){
                //if 'hide' event is triggered on this modal and not bubbled up from its child elements
                if( e.target === this.el ){ 
                    this.remove(); 
                }      
            }
        },
        handleSubmitButtonState: function(model) {
   
            if(this.model.report.get('elementCreateType') === 'pivot') {
                var pageInfo = utils.getPageInfo();
                this.$(Modal.FOOTER_SELECTOR)
                    .find('.btn-primary').replaceWith('<a href="'+route.pivot(pageInfo.root, pageInfo.locale, pageInfo.app)+'" class="btn btn-primary modal-btn-primary">' + _('Go to Pivot').t() + '</a>');
            }
            else{
                this.$(Modal.FOOTER_SELECTOR)
                    .find('.btn-primary').replaceWith('<a href="#" class="btn btn-primary modal-btn-primary">' + _('Add Panel').t() + '</a>');             
            }

        },
        setLabel: function() {
            var timeLabel = this.model.timeRange.generateLabel(this.collection);
            this.$el.find("span.time-label").text(timeLabel);
        },
        render: function() {
            this.$el.html(Modal.TEMPLATE);

            this.$(Modal.HEADER_TITLE_SELECTOR).html(_("Add Panel").t());

            this.$(Modal.BODY_SELECTOR).remove();

            this.$(Modal.FOOTER_SELECTOR).before(
                '<div class="vis-area">' +
                    '<div class="slide-area">' +
                        '<div class="content-wrapper add-panel-wrapper">' +
                            '<div class="' + Modal.BODY_CLASS + '" >' +
                            '</div>' +
                        '</div>' +
                        '<div class="timerange-picker-wrapper">' +
                        '</div>' +
                    '</div>' +
                '</div>'
            );

            this.$(Modal.BODY_SELECTOR).prepend(this.children.flashMessages.render().el);
            this.$(Modal.BODY_SELECTOR).append(Modal.FORM_HORIZONTAL_JUSTIFIED);

            this.$visArea = this.$('.vis-area').eq(0);
            this.$slideArea = this.$('.slide-area').eq(0);
            this.$addpanelContent = this.$('.add-panel-wrapper').eq(0);
            this.$timeRangePickerWrapper = this.$('.timerange-picker-wrapper').eq(0);

            this.$(Modal.BODY_FORM_SELECTOR).append(this.children.addPanel.render().el);
            this.$timeRangePickerWrapper.append(this.children.timeRangePickerView.render().el);

            this.$modalParent = this.$el;

            this.timeRangeDelegate = new TimeRangeDelegate({
                el: this.el,
                $visArea: this.$visArea,
                $slideArea: this.$slideArea,
                $contentWrapper: this.$addpanelContent,
                $timeRangePickerWrapper: this.$timeRangePickerWrapper,
                $modalParent: this.$modalParent,
                $timeRangePicker: this.children.timeRangePickerView.$el,
                activateSelector: 'a.timerange-control',
                backButtonSelector: 'a.btn.back'
            });

            this.$(Modal.FOOTER_SELECTOR).append(Modal.BUTTON_CANCEL);
            this.$(Modal.FOOTER_SELECTOR).append('<a href="#" class="btn btn-primary modal-btn-primary">' + _('Add Panel').t() + '</a>');

            this.$(Modal.FOOTER_SELECTOR).append('<a href="#" class="btn back modal-btn-back pull-left">' + _('Back').t() + '</a>');
            this.$('.btn.back').hide();

            return this;
        }
    });

});
