/**
 * @author jszeto
 * @date 11/22/13
 *
 * Controller class to kick off searches and manage the job lifecycle. This controller reuses a single Job model
 *
 *  model: {
 *      application {models/shared/Application}
 *      dataModel {models/services/datamodel/DataModel} the data model currently being edited
 *      sampleSize {models/Base} - contains one string attribute called "sampleSize". Used to restrict sample size of search
 *      searchJob {models/search/Job} - shared search job
 */
define([
    'jquery',
    'underscore',
    'backbone',
    'module',
    'controllers/Base'

],
    function (
        $,
        _,
        Backbone,
        module,
        BaseController
        ) {

        return BaseController.extend({
            moduleId: module.id,
            ADHOC_SEARCH_LEVEL: "verbose", // TODO [JCS] Parameterize this because some jobs don't need summary info

            initialize: function(options) {
                BaseController.prototype.initialize.call(this, options);

                this.model.searchJob.on('change:id', function() {
                    if (this.model.searchJob.isNew()) {
                        this.model.searchJob.fetchAbort();
                        this.model.searchJob.stopPolling();
                        this.model.searchJob.stopKeepAlive();
                    } else {
                        this.model.searchJob.startPolling();
                    }
                }, this);
            },

            createJobAttributes: function(searchString, earliest, latest, head) {
                var result = {
                    search: searchString  || '',
                    earliest_time: earliest || null,
                    latest_time: latest || null,
                    auto_cancel: 100,
                    auto_finalize_ec: 0,
                    status_buckets: 0,
                    timeFormat: '%s.%Q',
                    ui_dispatch_app: this.model.application.get("app"),
                    rf: '*',
                    preview: true,
                    adhoc_search_level: this.ADHOC_SEARCH_LEVEL,
                    app: this.model.application.get("app"),
                    owner: this.model.application.get("owner")
                };

                if(head) {
                    result.search += (' | head ' + head);
                }

                return result;
            },

            previewBySearchString: function(searchString) {
                var value = this.model.sampleSize.get("sampleSize");

                this.startNewSearchJob(searchString, value.earliest, value.latest, value.head);
            },

            /**
             * Called when user presses the Preview button. Posts the changes to the backend with the provisional flag.
             * Then retrieves the search string and kicks off a search job
             *
             * @param {string} objectName
             * @param {array} calculationFieldNames - array of strings of the fields to preview
             */
            preview: function(objectName, calculationFieldNames) {
                var value = this.model.sampleSize.get("sampleSize");

                // TODO [JCS] Move the provisional object save logic up here from the DataModelEditor router
                this.fetchPreviewSearchString(objectName, calculationFieldNames).then(function(searchString) {
                    this.startNewSearchJob(searchString, value.earliest, value.latest, value.head);
                }.bind(this));
            },

            /**
             * Helper function to get the preview search string when adding or editing an Object.
             * We ask the backend to save a provisional copy of the dataModel and then ask it for the search string.
             *
             * @param {string} objectName - Name of the object that has been modified
             * @param {array} calculationFieldNames - Array of the fieldNames of the calculation's outputFields. If undefined,
             * then we just use previewSearch
             * @return {Deferred} - Returns a Deferred object that is done after the dataModel has been saved.
             */
            fetchPreviewSearchString: function(objectName, calculationFieldNames) {
                var dfd = $.Deferred();

                this.model.dataModel.save({}, {data: {provisional: true}}).then(function() {
                    var objectModel = this.model.dataModel.objectByName(objectName);
                    var searchString;

                    if (!_(calculationFieldNames).isUndefined()) {
                        searchString = objectModel.getCalculationPreviewSearch(calculationFieldNames, true, true);
                    } else {
                        searchString = objectModel.get("previewSearch");
                    }

                    dfd.resolve(searchString);
                }.bind(this));

                return dfd;
            },

            /**
             * This function will take any search string and run a search job. It will then display an events view or
             * table view depending upon the type of search.
             *
             * @param {string} searchString - string to use for the search
             * @param {string} earliest - earliest time for the search
             * @param {string} latest - latest time for the search
             * @param {number} head - number of results to return from head
             */
            startNewSearchJob: function(searchString, earliest, latest, head) {
                // Cleanup any existing search jobs
                this.resetSearchJob();
                // Run the search job and start polling so we get progress updates
                this.model.searchJob.save({}, { data: this.createJobAttributes(searchString, earliest, latest, head) });
            },

            resetSearchJob: function() {
                this.model.searchJob.clear();
            }
        });

    });

