/**
 * @author jszeto
 * @date 12/6/13
 *
 * Wrapper view with a results table subview. If the results table has no results, then we show an error message instead.
 * The error message is passed in by the parent since it the wording is usually context sensitive.
 *
 * Inputs:
 *
 *  resultsTableClass {View} The class to use to generate the results table. Default is views/shared/results_table/ResultsTableMaster
 *  tableOptions {Object} The options to pass to the results table upon creation.
 *  noResultsMessage {Object}
 *      type {string} The type of message (splunkDUtils.ERROR | splunkDUtils.WARNING | splunkDUtils.INFO)
 *      html {string} The message text
 *  model {Object} The model object to pass to the resultsTableClass. It must contain a results model.
 *      results {models/services/search/jobs/ResultJsonRows} contains the rows and fields for the results
 *
 *
 */
define([
    'jquery',
    'underscore',
    'backbone',
    'module',
    'views/Base',
    'views/shared/results_table/ResultsTableMaster',
    'views/shared/FlashMessages',
    'util/splunkd_utils'
],
    function (
        $,
        _,
        Backbone,
        module,
        BaseView,
        ResultsTableMaster,
        FlashMessagesView,
        splunkdUtils
        ) {

        var ResultsTableContainer = BaseView.extend({
            moduleId: module.id,

            initialize: function(options) {
                BaseView.prototype.initialize.call(this, options);

                _(options).defaults({resultsTableClass: ResultsTableMaster, tableOptions: {}});
                _(options.tableOptions).defaults({enableTableDock: true, className: "scrolling-table-wrapper"});
                options.tableOptions.model = this.model;

                this.model.results.on("change:rows", this.resultsRowsChangeHandler, this);

                // Create the results table
                this.children.resultsTable = new options.resultsTableClass(options.tableOptions);

                this.children.resultsTable.on('cellClick', function(rowInfo) {
                    this.trigger('cellClick', rowInfo);
                }, this); 

                this.children.flashMessagesView = new FlashMessagesView({model:[this.model.results], whitelist: [splunkdUtils.ERROR]});
            },

            resultsRowsChangeHandler: function() {
                var rows = this.model.results.get("rows");
                if (rows.length == 0) {
                    this.children.flashMessagesView.flashMsgHelper.addGeneralMessage(ResultsTableContainer.NO_RESULTS_ID,
                        this.options.noResultsMessage);
                } else {
                    this.children.flashMessagesView.flashMsgHelper.removeGeneralMessage(ResultsTableContainer.NO_RESULTS_ID);
                }
            },

            render: function() {
                // Detach children

                // Use template
                this.$el.html(this.compiledTemplate({}));

                // Attach children and render them
                this.children.resultsTable.render().appendTo(this.$(".results-table-placeholder"));
                this.children.flashMessagesView.render().appendTo(this.$(".flash-messages-placeholder"));
                return this;
            },

            template: '\
                <div class="results-table-placeholder"></div>\
                <div class="flash-messages-placeholder" style="height:100%;"></div>\
            '
        },
        {
            NO_RESULTS_ID : "__NO_RESULTS_ID__",
            NO_RESULTS_MSG : "No results found."

        });

        return ResultsTableContainer;

    });

