/**
 * @author jszeto
 * @date 11/6/13
 *
 * Preview view for Add / Edit Rex Calculation. It contains three PreviewResultsPane views to display all events,
 * matching events and non-matching events. It also has a ViewValuesPane for each output field.
 *
 * The panes use the same Job model. Each has its own ResultJsonRows and ResultsFetchData models. The ResultsFetchData
 * models are given the count, offset, sort info and sub search.
 *
 * The output field panes use the Job summary model
 *
 * Inputs:
 *
 *  model: {
 *      searchJob {models/search/Job} - shared search job
 *      application {models/shared/Application}
 *      dataModel {models/services/datamodel/DataModel} the data model currently being edited
        sampleSize {models/Base} - contains one string attribute called "sampleSize". Used to restrict sample size of search
 *  }
 */
define([
    'jquery',
    'underscore',
    'backbone',
    'module',
    'models/Base',
    'models/services/search/jobs/Summary',
    'models/shared/fetchdata/ResultsFetchData',
    'models/services/search/jobs/ResultJsonRows',
    'views/Base',
    'views/shared/tabcontrols/TabbedViewStack',
    'views/shared/results_table/MatchResultsTableMaster',
    './PreviewResultsPane',
    './ViewValuesPane',
    './DrillDownResultsTable',
    'util/splunkd_utils',
    'util/datamodel/search_utils'
],
    function (
        $,
        _,
        Backbone,
        module,
        BaseModel,
        SummaryModel,
        ResultsFetchData,
        ResultJsonRows,
        BaseView,
        TabbedViewStack,
        MatchResultsTableMaster,
        PreviewResultsPane,
        ViewValuesPane,
        DrillDownResultsTable,
        splunkdUtils,
        search_utils) {

        var RexPreviewViewStack = BaseView.extend({
            moduleId:module.id,


            initialize:function (options) {
                BaseView.prototype.initialize.call(this, options);
                this.model.searchJob.on("prepared", this.searchJobPrepared, this);
                this.model.searchJob.on('change:id', this.searchJobClear, this);

                this.model.summary = new SummaryModel();
                this.model.searchJob.entry.links.on('change', function() {
                    this.model.summary.set("id", this.model.searchJob.entry.links.get('summary'));
                },this);
            },

            createSubViews: function() {
                // Create the TabbedViewStack and the three View panes (All, Match, Non-match)
                var calculationModel = this.model.calculation;
                var objectModel = this.model.object;
                var inputField = calculationModel.get("inputField");

                if (this.allPane) {
                    this.allPane.remove();
                }
                if (this.matchPane) {
                    this.matchPane.remove();
                }
                if (this.nonmatchPane) {
                    this.nonmatchPane.remove();
                }
                if (this.viewValuesPanes) {
                    _(this.viewValuesPanes).each(function(viewValuesPane) {
                        viewValuesPane.remove();
                    }, this);
                }
                if (this.children.viewStack) {
                    this.children.viewStack.remove();
                }

                // Calculate the search sub strings
                var allFields = _.union(["is_match", inputField],
                                    calculationModel.getFieldNames());
                var matchFields = _.union([inputField],
                    calculationModel.getFieldNames());
                var nonmatchFields = [inputField];

                var allSearch = calculationModel.getPreviewAllSearch() + search_utils.generateFieldSubSearch(allFields, inputField);
                var matchSearch = calculationModel.getPreviewMatchSearch() + search_utils.generateFieldSubSearch(matchFields, inputField);
                var nonmatchSearch = calculationModel.getPreviewNonMatchSearch() + search_utils.generateFieldSubSearch(nonmatchFields, inputField);

                // Create the fetch data for each pane, each with its own unique sub search
                this.model.allFetchData = new ResultsFetchData({output_mode: 'json_rows', segmentation:"full", count:20, max_lines: 20, search: allSearch});
                this.model.matchFetchData = new ResultsFetchData({output_mode: 'json_rows', segmentation:"full", count:20, max_lines: 20, search: matchSearch});
                this.model.nonmatchFetchData = new ResultsFetchData({output_mode: 'json_rows', segmentation:"full", count:20, max_lines: 20, search: nonmatchSearch});
                this.model.drillDownFetchData = new ResultsFetchData({output_mode: 'json_rows', segmentation:"full", count:20, max_lines: 20, search: ""});

                // Create the results models
                this.model.allResults = new ResultJsonRows({}, {fetchData: this.model.allFetchData});
                this.model.matchResults = new ResultJsonRows({}, {fetchData: this.model.matchFetchData});
                this.model.nonmatchResults = new ResultJsonRows({}, {fetchData: this.model.nonmatchFetchData});
                this.model.drillDownResults = new ResultJsonRows({}, {fetchData: this.model.drillDownFetchData});

                // Hookup the results model id to the search job links url
                this.model.allResults.set("id", this.model.searchJob.entry.links.get("results_preview"));
                this.model.matchResults.set("id", this.model.searchJob.entry.links.get("results_preview"));
                this.model.nonmatchResults.set("id", this.model.searchJob.entry.links.get("results_preview"));
                this.model.drillDownResults.set("id", this.model.searchJob.entry.links.get("results_preview"));

                // index=_internal status=404 rex field=uri http://(?<host>.*)/ max_match=1 | fields - _*, is_match, uri, host | eval is_match = if(match(uri, "http://(?<host>.*)/"), 1, 0)
                // index=_internal status=404 rex field=uri http://(?<host>.*)/ max_match=1 | fields - _*, uri, host | where match(uri, "http://(?<host>.*)/")
                // index=_internal status=404 rex field=uri http://(?<host>.*)/ max_match=1 | fields - _*, uri | where NOT match(uri, "http://(?<host>.*)/")

                // Create each of the preview panes
                this.allPane = new PreviewResultsPane({
                    model: {
                        dataModel: this.model.dataModel,
                        application: this.model.application,
                        sampleSize: this.model.sampleSize,
                        searchJob: this.model.searchJob,
                        resultsFetchData: this.model.allFetchData,
                        results: this.model.allResults},
                    flashMessagesView: this.options.flashMessagesView,
                    label:_("All").t(),
                    resultsTableMasterClass: MatchResultsTableMaster,
                    noResultsMessage: { type: splunkdUtils.ERROR,
                                        html: _("No results found.").t()}});
                this.matchPane = new PreviewResultsPane({
                    model: {
                        dataModel: this.model.dataModel,
                        application: this.model.application,
                        sampleSize: this.model.sampleSize,
                        searchJob: this.model.searchJob,
                        resultsFetchData: this.model.matchFetchData,
                        results: this.model.matchResults},
                    flashMessagesView: this.options.flashMessagesView,
                    label:_("Match").t(),
                    noResultsMessage: { type: splunkdUtils.ERROR,
                        html: _("None of the results match the regular expression.").t()}});
                this.nonmatchPane = new PreviewResultsPane({
                    model: {
                        dataModel: this.model.dataModel,
                        application: this.model.application,
                        sampleSize: this.model.sampleSize,
                        searchJob: this.model.searchJob,
                        resultsFetchData: this.model.nonmatchFetchData,
                        results: this.model.nonmatchResults},
                    flashMessagesView: this.options.flashMessagesView,
                    label:_("Non-Match").t(),
                    tabClassName: "tab-results-last",
                    noResultsMessage: { type: splunkdUtils.WARNING,
                        html: _("All results match the regular expression.").t()}});
                
                // For each output field, create a ViewValues Pane
                this.viewValuesPanes = [];
                this.model.calculation.withEachField(function(field) {
                    var valuesPane = new ViewValuesPane({label:field.get("displayName"),
                                                                 model: {
                                                                     field: field,
                                                                     sampleSize: this.model.sampleSize,
                                                                     searchJob: this.model.searchJob,
                                                                     summary: this.model.summary
                                                                 },
                                                                 noResultsMessage: { type: splunkdUtils.WARNING,
                         html: _("The attribute was not found in the events. Try changing the sample size or the regular expression.").t()}}); 
                    this.viewValuesPanes.push(valuesPane); 
                    
                    valuesPane.on('cellClick', this.handleCellClick, this); 

                }, this);

                // Combine all the panes into an array
                this.children.previewPanes = [this.allPane, this.matchPane, this.nonmatchPane].concat(this.viewValuesPanes);

                // Create array of visited panes
                this.visitedPanes = [];
                for (var i = 0; i < this.children.previewPanes.length; i++) {
                    this.visitedPanes[i] = false;
                }

                // Create the view stack to toggle between the panes
                this.children.viewStack = new TabbedViewStack({selectedIndex:0});

                this.children.viewStack.on("change:selectedIndex", this.selectedIndexChangeHandler, this);
                this.children.viewStack.render().appendTo(this.$(".viewstack-placeholder"));

                this.toggleViewStackPanes();
            },
            handleCellClick: function(rowInfo) {
                var calculationModel = this.model.calculation;
                var inputField = calculationModel.get("inputField");
                var drillDownFields = _.union(['_raw'], [inputField], calculationModel.getFieldNames());
                var drillDownSearch = "search " + rowInfo.field.id + '=' + '"' + rowInfo.value.replace(/"/g, '\\"') + '"' +  search_utils.generateFieldSubSearch(drillDownFields, inputField);
                this.model.drillDownFetchData.set('search', drillDownSearch);
                this.model.drillDownResults.safeFetch();

                this.model.rowInfo = new Backbone.Model();
                this.drillDownTable = new DrillDownResultsTable({
                        model: {
                            dataModel: this.model.dataModel,
                            searchJob: this.model.searchJob,
                            resultsFetchData: this.model.drillDownFetchData,
                            rowInfo: this.model.rowInfo,
                            results: this.model.drillDownResults
                        },
                        flashMessagesView: this.options.flashMessagesView,
                        label:_("Drill-down").t(),
                        tabClassName: "tab-results-last",
                        noResultsMessage: { type: splunkdUtils.WARNING,
                            html: _("None of the results match the selected row.").t()}
                });

                this.drillDownTable.on("backToRegexClicked", function() {
                        this.trigger('hideDrillDownView', this.drillDownTable);
                }, this);
                this.model.rowInfo.set('field', rowInfo.field.id);
                this.model.rowInfo.set('value', rowInfo.value);
                this.trigger('showDrillDownView', this.drillDownTable);
            }, 
            /**
             * Called by parent to pass in the current Object and Calculation models
             *
             * @param objectModel {models/services/datamodel/private/Object}
             * @param calculationModel {models/services/datamodel/private/Calculation}
             */
            setModels: function(objectModel, calculationModel) {
                this.model.object = objectModel;
                this.model.calculation = calculationModel;
            },

            /**
             * Job has been kicked off. Let's create all of the panes
             */
            searchJobPrepared: function() {
                this.previousHasEvents = undefined;
                this.model.searchJob.on("jobProgress", this.searchJobProgress, this);
                this.model.searchJob.entry.content.on("change:eventCount", this.toggleViewStackPanes, this);
                this.createSubViews();
            },

            /**
             * Job has been cleared. Let's reset our state
             */
            searchJobClear: function() {
                if (this.model.searchJob.isNew()) {
                    this.model.searchJob.off("jobProgress", this.searchJobProgress, this);
                    this.model.searchJob.entry.content.off("change:eventCount", this.toggleViewStackPanes, this);
                    this.model.summary.clear();
                }
            },

            /**
             * Job is making progress. Fetch the summary info and the results for the currently visible pane
             */
            searchJobProgress: function() {
                this.model.summary.safeFetch({data: {top_count:1000}});
                this.fetchDataForSelectedPane(this.children.viewStack.getSelectedIndex());
            },

            normalizeEventCount: function(value) {
                if (value == undefined)
                    return false;
                value = parseInt(value, 10);
                if (value == 0)
                    return false;
                return true;
            },

            toggleViewStackPanes: function() {
                var hasEvents = this.normalizeEventCount(this.model.searchJob.entry.content.get("eventCount"));
//                console.log("RexPreviewViewStack eventCount",this.model.searchJob.entry.content.get("eventCount"),"hasEvents",hasEvents,"previousHasEvents",this.previousHasEvents);
                if (hasEvents != this.previousHasEvents) {
                    var panesLength = this.children.viewStack.getLength();
                    if (hasEvents) {
                        if (panesLength <= 1) {
//                            console.log("RexPreviewViewStack.toggleViewStackPanes adding panes");
                            var length = this.children.previewPanes.length;
                            for (var j=panesLength; j < length; j++) {
                                var pane = this.children.previewPanes[j];
                                this.children.viewStack.addPane(pane);
                            }
                        }
                    } else {
                        if (panesLength > 1) {
//                            console.log("RexPreviewViewStack.toggleViewStackPanes removing panes");
                            // Remove all but the first pane
                            for (var i = panesLength - 1; i > 0; i--) {
                                this.children.viewStack.removePaneAt(i);
                            }
                        } else if (panesLength == 0) {
//                            console.log("RexPreviewViewStack.toggleViewStackPanes adding first pane");
                            this.children.viewStack.addPane(this.children.previewPanes[0]);
                        }

                    }
                }
                this.previousHasEvents = hasEvents;
            },


            /**
             * The user has navigated to a new pane. Fetch the results for the pane if needed.
             */
            selectedIndexChangeHandler: function() {
                var selectedIndex = this.children.viewStack.getSelectedIndex();
                // Only fetch if job is still running or we haven't visited the pane yet
                if (this.model.searchJob.isRunning() || !this.visitedPanes[selectedIndex]) {
                    this.fetchDataForSelectedPane(selectedIndex);
                }

                this.visitedPanes[selectedIndex] = true;
            },

            fetchDataForSelectedPane: function(selectedIndex) {
                if (selectedIndex == 0)
                    this.model.allResults.safeFetch();
                else if (selectedIndex == 1)
                    this.model.matchResults.safeFetch();
                else if (selectedIndex == 2)
                    this.model.nonmatchResults.safeFetch();
            },

            render:function () {
                // Detach children

                // Use template
                this.$el.html(this.compiledTemplate({}));
                // Attach children and render them

                return this;
            },

            template:'\
                <div class="viewstack-placeholder"></div>\
            '
        },
        {
            SUMMARY_TOP_COUNT:1000
        });

        return RexPreviewViewStack;

    });

