/**
 * @author jszeto
 * @date 11/9/12
 *
 * Base class for the form subviews for Adding or Editing Objects, Transactions or Calculations
 *
 * Inputs:
 *
 *     model: {
 *        dataModel {models/services/datamodel/DataModel} the data model currently being edited
 *        appLocal {models/services/AppLocal}
 *        application {models/Application}
 *        buttonSettings {base/Model}
 *    }
 *     collection { [optional]
 *        transformsLookups {collections/services/data/TransformsLookups} - Collection of lookups to use
 *              when add/edit a Lookup calculation
 *     objectName {string}
 *     operation {string} (add|edit)
 *     type {string} (event|child|transaction|search|Eval|Lookup|Rex|GeoIP)
 *     parentObjectName {string} [optional] The name of the parent Object to either a child Object or a calculation
 *     calculationID {string} [optional]
 */
define([
    'jquery',
    'underscore',
    'views/ValidatingView',
    'util/splunkd_utils',
    'splunk.util'
],
    function(
        $,
        _,
        ValidatingView,
        splunkDUtils,
        splunkUtils
        )
    {
        return ValidatingView.extend({

            /**
             * {DataModel} - Model for the data model being edited
             */
            dataModel: undefined,
            /**
             * {String} - The name of the object to edit
             */
            objectName: "",
            /**
             * {String} - The name of the parent Object to either a child Object or a calculation (lookup|regex|eval)
             */
            parentObjectName: "",
            /**
             * {String} - Whether to add or edit. Possible values = [add|edit]
             */
            operation: "add",
            /**
             * {String} - (Optional) ID of the Calculation
             */
            calculationID : "",
            /**
             * The FlashMessagesHelper created in the AddEditContainerView
             */
            flashMessagesHelper: undefined,

            initialize: function(options) {
                ValidatingView.prototype.initialize.call(this, options);
                options = options || {};
                this.objectName = options.objectName;
                this.parentObjectName = options.parentObjectName;
                this.operation = options.operation;
                this.calculationID = options.calculationID;
                this.flashMessagesHelper = options.flashMessagesHelper;
            },
            /**
             * Tells the view to save the inputs into the dataModel. resultCallback will be called after we get a response
             * from the splunkd
             *
             * @param resultCallback {Function}
             */
            save: function(resultCallback) {

            },
            /**
             * Tells the view to provisionally save the inputs into the dataModel. resultCallback will be called after we
             * get a response from splunkd.
             *
             * @param resultCallback {Function}
             */
            preview: function(resultCallback) {

            },
            /**
             * Tells the view to perform unbinding and any other cleanup before it is destroyed.
             */
            cleanup: function() {

            },

            render: function() {
                var errorCount = this.flashMessagesHelper.getGeneralMessagesSize();
                if (errorCount == 0)
                    this.renderContent();
            },

            /**
             * Helper function to hide the Save button
             */
            disableSave: function() {
                this.model.buttonSettings.set({save:"hide",preview:"hide","delete":"hide"});
            },

            /**
             * Display an error about not finding the Object
             * @param objectName
             */
            addObjectError: function(objectName) {
                var msg = splunkUtils.sprintf(_("Could not find the Object <i>%s</i>").t(), objectName);

                this.flashMessagesHelper.addGeneralMessage(this.cid + "_missingObject",
                    { type: splunkDUtils.ERROR,
                        html: msg});

                this.disableSave();
            },

            /**
             * Display an error about not finding the Calculation
             * @param calculationID
             */
            addCalculationError: function(calculationID) {
                var msg = splunkUtils.sprintf(_("Could not find the Calculation <i>%s</i>").t(), calculationID);
                this.flashMessagesHelper.addGeneralMessage(this.cid + "_missingCalculation",
                    { type: splunkDUtils.ERROR,
                        html: msg});

                this.disableSave();
            },

            /**
             * Display an error about not finding the Field
             * @param fieldName
             */
            addFieldError: function(fieldName) {
                var msg = splunkUtils.sprintf(_("Could not find the Field <i>%s</i>").t(), fieldName);
                this.flashMessagesHelper.addGeneralMessage(this.cid + "_missingField",
                    { type: splunkDUtils.ERROR,
                        html: msg});

                this.disableSave();
            }
        });

    });