/**
 * @author jszeto
 * @date 11/21/12
 *
 *  *  inputs:
 *      model
 *          calculation {models/services/datamodel/private/RexCalculation} - The rex calculation to edit
 *          objectModel {models/services/datamodel/private/Objects} - The object model that owns the calculation model
 *          dataModel {models/services/datamodel/DataModel} - The data model that owns the object model
 *
 * A subview to add or edit a Rex Calculation.
 *
 */

define([
    'jquery',
    'underscore',
    'views/data_model_editor/form_views/AddEditFormBaseView',
    'views/shared/controls/ControlGroup',
    'views/data_model_editor/form_components/TrimmedTextareaControl',
    'views/data_model_editor/form_components/RexOutputFieldsView',
    'uri/route',
    'util/splunkd_utils',
    'splunk.util',
    'module'
],
    function(
        $,
        _,
        AddEditFormBaseView,
        ControlGroup,
        TrimmedTextareaControl,
        RexOutputFieldsView,
        route,
        splunkDUtils,
        splunkUtils,
        module
        )
    {
        return AddEditFormBaseView.extend({
            moduleId: module.id,
            className: 'rex-form-view',
            MISSING_OUTPUTFIELD_ID: "MISSING_OUTPUTFIELD_ID",

            // TODO [JCS] Remove
            modelToControlGroupMap:  {
                inputField: "selectField",
                expression: "textareaExpression"
            },

            /**
             * @constructor
             * @param options {Object} {
                 }
             */
            initialize: function(options) {
                AddEditFormBaseView.prototype.initialize.call(this, options);

                this.model.objectModel = this.model.dataModel.objectByName(this.parentObjectName);

                if (_(this.model.objectModel).isUndefined()) {
                    this.addObjectError(this.parentObjectName);
                    return;
                }

                // SyntheticSelect objects for input fields
                var fieldNames = _(this.model.objectModel.getAllowedInputFields()).map(function(field) {
                        return {label: field.displayName, value:field.fieldName};
                    }
                );

                // TODO [JCS] If fieldNames is empty, don't allow any adding and show an error message
                if (this.operation == "add")
                {
                    // Default inputField to the first field
                    var calcAttributes = {calculationType: "Rex",
                                          inputField: fieldNames[0].value,
                                          calculationID: "",
                                          expression:""};
                    // Create a new Calculation model
                    this.model.calculation = this.model.objectModel.createCalculation(calcAttributes);
                }
                else
                {
                    this.model.calculation = this.model.objectModel.getCalculation(this.calculationID);

                    if (_(this.model.calculation).isUndefined()) {
                        this.addCalculationError(this.calculationID);
                        return;
                    }
                }

                // Select the input field
                this.children.selectField = new ControlGroup({label: _("Extract From").t(),
                                                              controlType: "SyntheticSelect",
                                                              controlOptions:
                                                                    {model:this.model.calculation,
                                                                     modelAttribute:"inputField",
                                                                     items:fieldNames,
                                                                     toggleClassName: 'btn'}});

                var textareaChild = new TrimmedTextareaControl({model:this.model.calculation,
                                                         modelAttribute:"expression"});
                this.children.textareaExpression = new ControlGroup({label: _("Regular Expression").t(),
                                                                     controlType: "Textarea",
                                                                     controls: textareaChild});
                // TODO [JCS] Listen on keyUp, not change (aka blur)
                textareaChild.on("change", this.textareaExpressionChangeHandler, this);

                this.children.outputFields = new RexOutputFieldsView({model: this.model.calculation});

                this.flashMessagesHelper.register(this.model.calculation);
            },

            /**
             * Called whenever the Expression textarea value has changed
             * @param {string} new value for the textarea
             */
            textareaExpressionChangeHandler: function(value) {
                // Find all of the field names and update the calculation's outputFields
                // As a rendering optimization, we only update the model for the fields that have been added or removed.

                value = splunkUtils.trim(value);
                // Find all of the capture groups ( the text btw <>) in the user's regular expression. Capture groups
                // map to outputField fieldNames.
                var re = new RegExp("<[^>]*>", "gi");
                var matches = value.match(re);

                if (!_(matches).isEmpty())
                {
                    var fieldsToRemove = [];
                    // Strip the < and > from the matches
                    matches = _(matches).map(function(match) {
                       return match.substring(1, match.length - 1);
                    });

                    // Delete any fieldNames that don't have a match
                    this.model.calculation.withEachField(function(field) {
                        var fieldName = field.get("fieldName");
                        if (!_.isUndefined(fieldName))
                        {
                            var matchedFieldName = _(matches).find(function(match) {
                                return fieldName == match;
                            });
                            if (_.isUndefined(matchedFieldName)) {
                                fieldsToRemove.push(fieldName);
                            }
                        }
                    }, this);

                    // Delete any fields marked for removal
                    _(fieldsToRemove).each(function(fieldToRemove) {
                        var fieldObject = this.model.calculation.getFieldByName(fieldToRemove);
                        this.flashMessagesHelper.unregister(fieldObject);
                        this.model.calculation.removeFieldByName(fieldToRemove);
                    }, this);

                    // See if a field is already in our outputFields. If not, add a new outputField
                    _(matches).each(function(fieldName) {
                        var field = this.model.calculation.getFieldByName(fieldName);
                        if (_.isUndefined(field)) {
                            this.flashMessagesHelper.register(this.model.calculation.addField({fieldName:fieldName}));
                        }
                    }, this);
                } else {
                    this.model.calculation.withEachField(function(field) {
                        this.flashMessagesHelper.unregister(field);
                    }, this);
                    this.model.calculation.removeAllFields();
                }
            },

            /**
             * Called by both save and preview to perform validation on the models
             * @return {boolean}
             */
            applyControlsToModel: function() {

                var isErrorRexExp = false;
                var isValid = this.model.calculation.set({}, {validate:true});

                if (this.model.calculation.outputFields.length == 0) {
                    this.flashMessagesHelper.addGeneralMessage(this.MISSING_OUTPUTFIELD_ID,
                        {type: splunkDUtils.ERROR,
                            html: _("No attributes were extracted. Provide a working Regular Expression with capturing groups.").t()});
                    isValid = false;
                    isErrorRexExp = true;
                } else {
                    this.flashMessagesHelper.removeGeneralMessage(this.MISSING_OUTPUTFIELD_ID);
                }

                this.model.calculation.withEachField(function(field) {
                    isValid = field.set({}, {validate:true}) && isValid;
                    if (!isValid)
                        isErrorRexExp = true;
                }, this);

                // [JCS] For now we need to hack the textareaExpression error state since technically its model attribute
                // is not in an error state.
                if (isErrorRexExp) {
                    this.children.textareaExpression.error(true,_("Regular Expression is in a bad state").t());
                } else {
                    this.children.textareaExpression.error(false,"");
                }
                return isValid;
            },

            /**
             * Tells the view to save the inputs into the dataModel. resultCallback will be called after we get a response
             * from the splunkd
             *
             * @param resultCallback {Function}
             */
            save: function(resultCallback) {
                AddEditFormBaseView.prototype.save.call(this, resultCallback);

                if (this.applyControlsToModel()) {
                    var resultXHR = this.model.dataModel.save({});
                    if (resultXHR) {
                        resultXHR.done(_(function() {
                            this.trigger("action:save");
                        }).bind(this));
                    }
                }
            },

            /**
             * Tells the view to provisionally save the inputs into the dataModel. resultCallback will be called after we
             * get a response from splunkd.
             *
             * @param resultCallback {Function}
             */
            preview: function(resultCallback) {
                AddEditFormBaseView.prototype.preview.call(this, resultCallback);

                if (this.applyControlsToModel()) {
                    this.trigger("action:preview",
                        this.model.objectModel.get("objectName"),
                        _.union(["is_match",
                                 this.model.calculation.get("inputField")],
                                 this.model.calculation.getFieldNames()),
                        this.model.calculation);
                }
            },
            /**
             * Tells the view to perform unbinding and any other cleanup before it is destroyed.
             */
            cleanup: function() {
                AddEditFormBaseView.prototype.cleanup.call(this);
            },

            renderContent: function() {
                this.children.selectField.detach();
                this.children.textareaExpression.detach();

                var docUrl = route.docHelp(
                    this.model.application.get("root"),
                    this.model.application.get("locale"),
                    'search_app.assist.rex'
                );

                var html = _(this.template).template({
                    type: this.type,
                    operation: this.operation,
                    docUrl: docUrl,
                    hasAttributes: this.model.calculation.outputFields.length
                });
                this.$el.html(html);

                this.$(".selectField-placeholder").replaceWith(this.children.selectField.render().el);
                this.$(".rex-expression-placeholder").replaceWith(this.children.textareaExpression.render().el);
                this.$(".output-fields-placeholder").append(this.children.outputFields.render().el);

                return this;
            },

            template: '\
                <form>\
                    <div class="col-1">\
                        <div class="selectField-placeholder"></div>\
                    </div>\
                    <div class="col-2">\
                            <div class="rex-expression-placeholder"></div>\
                            <span class="help-block"><%- _("Examples:").t() %><br>\
                                <span class="mono-space">http://&lt;uri_domain&gt;/&lt;uri_query_string&gt;?&lt;uri_query&gt;<br>\
                                From: (?&lt;from&gt;.*) To: (?&lt;to&gt;.*)</span>\
                            </span>\
                            <a href="<%- docUrl %>" target="_blank" class="help-link external"><%- _("Learn More").t() %></a>\
                    </div>\
                    <div class="col-3">\
                        <label><%- _("Attribute(s)").t() %></label>\
                        <div class="output-fields-placeholder"></div>\
                    </div>\
                </form>\
            '

        });

    });

