define([
    'jquery',
    'underscore',
    'backbone',
    'module',
    'views/shared/controls/Control',
    'views/shared/delegates/Popdown',
    'util/math_utils',
    'bootstrap.tooltip'
],
function(
    $,
    _,
    Backbone,
    module,
    Control,
    Popdown,
    math_utils,
    tooltip
){
    /**
     * Synthetic Select dropdown a-la Bootstrap
     *
     * @param {Object} options
     *                        {Object} model The model to operate on
     *                        {String} modelAttribute The attribute on the model to observe and update on selection
     *                        {Object} items An array of elements having the following keys:
     *                                       label (textual display),
     *                                       value (value to store in model)
     *                                       icon (icon name to show in menu and button label)
     *                                       iconURL (URL of icon to show in menu and button label)
     *                                       enabled (optional boolean, defaults to true, whether to enable the selection)
     *                                       description (additional text)
     *                                       (ie, {label: 'Foo Bar', value: 'foo', icon: 'bar'}).
     *                        {Object} groupedItems Optionally use if you want optgroup style dropdowns
     *                                An array of arrays of items
     *                                label (textual display),
     *                                items an array of items having the following keys:
     *                                     label (textual display),
     *                                     value (value to store in model)
     *                                     icon (icon name to show in menu and button label)
     *                                     iconURL (URL of icon to show in menu and button label)
     *                                     enabled (optional boolean, defaults to true, whether to enable the selection)
     *                                     description (additional text)
     *                                     (ie, {label: 'Foo Bar', value: 'foo', icon: 'bar'}).
     *                        {String} prompt (Optional) String to display if the value is undefined. If prompt is undefined,
     *                                                   then the control will just display the first item. Default value
     *                                                   is undefined
     *                        {String} help (Optional) Html to display in the bottom help section
     *                        {String} label (Optional) Html to display as the button label
     *                        {String} toggleClassName (Optional) Class attribute to add to the parent element
     *                        {String} menuClassName (Optional) Class attribute to add to the dialog element
     *                        {String} menuWidth (Optional) narrow, normal, or wide
     *                        {String} additionalClassNames (Optional) Class attribute(s) to add to control
     *                        {String} iconClassName (Optional) Class attribute(s) to add to icon
     *                        {String} iconURLClassName (Optional) Class attribute(s) to add to iconURL
     *                        {Boolean} nearestValue (Optional) if true: try to select the nearest value from items for the value of the modelAttribute 
     *                        {String} descriptionPosition (Optional) [right|bottom] Default value is "right". Position of the element's description text
     *                        relative to the item's label
     */

    return Control.extend({
        moduleId: module.id,
        className: 'control btn-group',
        keys: {
           UP_ARROW: 38,
           DOWN_ARROW: 40
        },
        items: undefined,
        groupedItems: undefined,
        renderList: true,
        initialize: function() {
            var defaults = {
                toggleClassName: '' ,
                menuClassName: '',
                iconClassName: 'icon-large',
                iconURLClassName: 'icon-large',
                descriptionPosition: 'right',
                label: '',
                popdownOptions: {el: this.el},
                html: '',
                nearestValue: false,
                prompt: undefined
            };

            _.defaults(this.options, defaults);
            _.defaults(this.options.popdownOptions, defaults.popdownOptions);

            this.normalizeItems(this.options.items, this.options.groupedItems);

            var modelAttribute = this.options.modelAttribute;
            if (modelAttribute) {
                this.$el.attr('data-name', modelAttribute);
            }
            
            Control.prototype.initialize.apply(this, arguments);
            
        },
        
        activate: function(options) {
            options = options || {};
            if (this.active) {
                return Control.prototype.activate.apply(this, arguments);
            }
            this.renderList = true;
            this.items = undefined;
            this.groupedItems = undefined;
            this.normalizeItems(this.options.items, this.options.groupedItems);
            Control.prototype.activate.apply(this, arguments);
        },
        
        /**
         * Change the items and render the control using the new items
         *
         * @param items An array of items having the following keys:
         *                                label (textual display),
         *                                value (value to store in model)
         *                                icon (icon name to show in menu and button label)
         *                                enabled (optional boolean, defaults to true, whether to enable the selection)
         *                                (ie, {label: 'Foo Bar', value: 'foo', icon: 'bar'}).
         */
        setItems: function(items) {
            this.normalizeItems(items, undefined);
            this.debouncedRender();
        },

        /**
         * Change the groupedItems and render the control using the new groupedItems
         *
         * @param groupedItems An array of arrays of items
         *                                label (textual display),
         *                                items an array of items having the following keys:
         *                                     label (textual display),
         *                                     value (value to store in model)
         *                                     icon (icon name to show in menu and button label)
         *                                     enabled (optional boolean, defaults to true, whether to enable the selection)
         *                                     (ie, {label: 'Foo Bar', value: 'foo', icon: 'bar'}).
         */
        setGroupedItems: function(groupedItems) {
            this.normalizeItems(undefined, groupedItems);
            this.debouncedRender();
        },
        /**
         * GroupedItems are used for rendering while items are used for iteration. This function ensures that both
         * groupedItems and items are normalized to the correct format. This should only be called internally.
         *
         * @param items
         * @param groupedItems
         */
        normalizeItems: function(items, groupedItems) {
            if (!items && groupedItems){
                this.items = [];
                this.groupedItems = groupedItems.slice();

                if(!(groupedItems instanceof Array) ||
                    !(groupedItems[0].items instanceof Array) ||
                    _.isUndefined(groupedItems[0].items[0].value)){
                    throw new Error("Invalid groupedItems Input");
                }

                _.each(groupedItems, function(value, key){
                    _.each(value.items, function(item, index){
                        this.items.push(item);
                    }, this);
                }, this);
            } else if (items) {
                this.items = items.slice();
                this.groupedItems = [];
                var group = {};
                group.label = '';
                group.items = this.items;
                this.groupedItems.push(group);
            }
            
            if (this.options.nearestValue) {
                this.valuesAsInts = [];
                _.each(this.items, function(item){
                    var convertedToInt = parseInt(item.value, 10);
                    if (_.isNaN(convertedToInt)) {
                        throw new Error('You cannot use the nearestValue option with a SyntheticSelect control that has values other than ints!');
                    }
                    this.valuesAsInts.push(convertedToInt);
                }, this);
            }

            this.renderList = true;
        },
        stopListening: function() {
            if (arguments.length === 0) {
                $(window).off('.' + this.cid);
                this.$menu && this.$menu.off('.' + this.cid);                
            }
            Control.prototype.stopListening.apply(this, arguments);
        },
        show: function(){
            this.$menu.find(".icon-check:visible").closest('a').focus();
            this.$menu.on('click.' + this.cid, 'a', this.click.bind(this));
            $(window).on('keydown.' + this.cid, this.keydown.bind(this));
            this.trigger('popdownShown');
        },
        hide: function(){
            $(window).off('.' + this.cid);
            this.$menu.off('.' + this.cid);
            this.trigger('popdownHidden');
        },
        disable: function(){
            this.options.enabled = false;
            this.$('a.dropdown-toggle').addClass('disabled');
        },
        enable: function(){
            this.options.enabled = true;
            this.$('a.dropdown-toggle').removeClass('disabled');
        },
        tooltip: function(options){
            this.$('a.dropdown-toggle').tooltip(options);
        },
        click: function(e) {
            var $currentTarget = $(e.currentTarget);
            if (!$currentTarget.hasClass('disabled')) {
                this.setValue($currentTarget.data('value'), true);
                this.$(".dropdown-toggle").focus();
            }
            e.preventDefault();
        },
        keydown: function (e) {
            if (e.shiftKey || e.ctrlKey || e.metaKey || e.altKey)  {
                return true;
            }

            var $focused = $(':focus');
            var toggleFocused = false;
            var selectedItem = -1; 
            var optionLength = -1;
            var focusValue = $focused.data("value");
            var attachedTo = this.options.popdownOptions.attachDialogTo ? $(this.options.popdownOptions.attachDialogTo)[0] : this.el;

            if ($.contains(attachedTo, $focused[0])) {
                _.each(this.items, function(element, key){
                    element['value'] == focusValue ? selectedItem = key : false;
                }, this);
            }
            
            if (e.keyCode === this.keys['DOWN_ARROW']) {
                if (selectedItem < this.items.length -1 ) {
                    this.$menu.find("li:eq(" + (selectedItem + 1) + ") > a").focus();
                    return false;
                }
                e.preventDefault();
            } else if (e.keyCode === this.keys['UP_ARROW']) {
                if (selectedItem > 0 ) {
                    this.$menu.find("li:eq(" + (selectedItem - 1) + ") > a").focus();
                    return false;
                } else {
                    this.$(".dropdown-toggle").focus();
                    return false;
                }
            } else if (
                        ((e.keyCode >= 48) && (e.keyCode <= 57))  || //number
                        ((e.keyCode >= 65) && (e.keyCode <= 90))  || //uppercase letter
                        ((e.keyCode >= 97) && (e.keyCode <= 122)) || //lowercase letter
                        ((e.keyCode >= 128) && (e.keyCode <= 165))   //extended letter
                    ) {
                    
                var found = false;
                this.$menu.find(" a").each(function() {
                    if (found) {return;}
                
                    var option = $(this);
                    var firstChar = option.text().replace(/^\s\s*/, '').substring(0, 1).toLowerCase(); 
                    var keyChar = String.fromCharCode(e.keyCode).toLowerCase();                      
                    if (keyChar == firstChar) {
                        option.focus();
                    }
                });
                
            }             
        },

        findItem: function() {
            if (!this.items || !this.items.length) {
                return void(0);
            }

            // If the value is undefined and we have a prompt, then return the prompt
            if (_(this._value).isUndefined() && !_(this.options.prompt).isUndefined()) {
                return {label: this.options.prompt, value: undefined};
            }
            
            if (this.options.nearestValue) {
                var nearestObject = math_utils.nearestMatchAndIndexInArray(this._value, this.valuesAsInts);
                if (nearestObject.index != void(0)) {
                    return this.items[nearestObject.index];
                }
            }
            
            return _.find(this.items, function(element) {
                return _.isEqual(this._value, element.value);
            }, this) || this.items[0];
        },
        render: function() {
            //TODO [JCS] Check why the dropdown doesn't open after setItems is called.
            var groupedItems = this.groupedItems;
            var items = this.items;

            if (!items || !items.length) {
                return this;
            }

            var item = this.findItem();

            if (this.renderList) {
                var template = this.compiledTemplate({
                        groupedItems: groupedItems,
                        item: item,
                        options: this.options,
                        menuWidthClass: (this.options.menuWidth != "normal" ? "dropdown-menu-" + this.options.menuWidth : '')
                    });
                this.$el.html(template);
                
                //bind that values to the items
                this.$el.find('.dropdown-menu a').each(function(i, el) {
                    $(el).data('value', items[i].value);
                });
                
                this.$menu =  this.$('.dropdown-menu');

                if (this.children.popdown) {
                    this.children.popdown.remove();
                    this.stopListening(this.children.popdown);
                    delete this.children.popdown;
                }

                this.children.popdown = new Popdown($.extend(true, {}, this.options.popdownOptions));
                this.listenTo(this.children.popdown, "shown", this.show);
                this.listenTo(this.children.popdown, "hidden", this.hide);

                this.renderList = false;
            }

            var additionalClassNames = this.options.additionalClassNames;
            if(additionalClassNames) {
                this.$el.addClass(additionalClassNames); 
            }

            //Hide or show the checkmarks
            this.$menu.find('a').each(function(i, el) {
                var $el = $(el);
                $el.find('.icon-check')[_.isEqual($el.data('value'), item.value) ? 'show' : 'hide']();
            });    
                                    
            //Update the toggle label
            if (item.label) {
                this.$(".dropdown-toggle > .link-label").text(_(this.options.label).t() + " " + _(item.label).t());
            }
            this.$(".dropdown-toggle > i").attr('class',  "icon-" + item.icon);

            return this;
        },
        template: '\
            <a class="dropdown-toggle <%- options.toggleClassName %>" href="#">\
                <% if (item.icon) { %> <i class="icon-<%-item.icon%> icon-large"></i><% } %>\
                <span class="link-label"><%- options.label %> <%- item.label %></span><span class="caret"></span>\
            </a>\
            <div class="dropdown-menu dropdown-menu-selectable <%- options.menuClassName %> <%- menuWidthClass %> ">\
            <div class="arrow"></div>\
                <% _.each(groupedItems, function(group, key, fullItems) { %>\
                <ul>\
                <% _.each(group.items, function(element, index, list) { %>\
                    <li><a class="synthetic-select <%- element.enabled === false ? \"disabled\" : \"\"%>" href="#" data-value="<%- element.value %>">\
                        <i class="icon-check"></i>\
                        <% if (element.icon) { %> <i class="icon-<%-element.icon%> <%-options.iconClassName %>"></i><% } %>\
                        <% if (element.iconURL) { %> <img class="<%-options.iconURLClassName %>" src="<%-element.iconURL%>" alt="icon"><% } %>\
                        <span class="link-label"><%- element.label %></span>\
                        <% if (element.description && (options.descriptionPosition == "right")) { %> <span class="link-description"><%- element.description %></span><% } %>\
                        <% if (element.description && (options.descriptionPosition == "bottom")) { %> <span class="link-description-below"><%- element.description %></span><% } %>\
                    </a></li>\
                <% }); %>\
                </ul>\
                <% }); %>\
            </div>\
        '
    });
});
