/**
 * @author ykou
 * @date 01/30/2014
 *
 * @description This is a component for time range setting. The core structure is 'a textfield + a unit drop-down list'.
 *              The input data would be automatically converted to seconds. For example, if user input '1 day', the data
 *              stored in the model would be 3600
 */

define(
	[
		'underscore',
		'models/Base',
		'views/Base',
		'views/shared/controls/Control',
		'views/shared/controls/SyntheticSelectControl',
		'views/shared/controls/TextControl',
		'module'
	], 
	function (
		_,
		BaseModel,
		BaseView,
		Control,
		SyntheticSelectControl,
		TextControl,
		module
	) {
        var YEAR = 31536000;
        var MONTH = 2592000;
        var WEEK = 604800;
        var DAY = 86400;
        var HOUR = 3600;
        var MINUTE = 60;
        var SECOND = 1;

		return Control.extend({
            moduleId: module.id,

            initialize: function () {
                // we need this mediator because we need to convert seconds to 'number unit' combination.
            	var timeRangeMediator = new BaseModel();

                // populate data to the input fields. The reason to use listener is that, in some cases the model is
                // fetched after this view is created (so that this.model.get() doesn't work).
                this.listenToOnce(this.model, 'change:'+this.options.modelAttribute+' reset:'+this.options.modelAttribute, function () {
                    this.convertTimeRange(this.model, timeRangeMediator);
                });

            	this.children.number = new TextControl({
            		model: timeRangeMediator,
            		modelAttribute: 'number'
            	});

            	this.children.unit = new SyntheticSelectControl({
            		model: timeRangeMediator,
            		modelAttribute: 'unit',
            		toggleClassName: 'btn',
            		items: [
            			{label: _('second(s)').t(), value: SECOND},
            			{label: _('minute(s)').t(), value: MINUTE},
            			{label: _('hour(s)').t(), value: HOUR},
            			{label: _('day(s)').t(), value: DAY},
            			{label: _('week(s)').t(), value: WEEK},
            			{label: _('month(s)').t(), value: MONTH},
            			{label: _('year(s)').t(), value: YEAR}
            		]
            	});

            	// TODO: validate 'number' and 'unit'
            	this.listenTo(timeRangeMediator, 'change:number change:unit reset:number reset:unit', function(){
            		var number = parseInt(timeRangeMediator.get('number'), 10) || 0;
            		var unit = parseInt(timeRangeMediator.get('unit'), 10) || 1;
            		this.model.set(this.options.modelAttribute, number * unit);
                });
            },

            /**
             * convert time range from seconds to other units.
             * @param  {model} timeRangeModel    this.model
             * @param  {model} timeRangeMediator the mediator
             * @return {[type]}                   [description]
             */
            convertTimeRange: function (timeRangeModel, timeRangeMediator) {
            	var timeRange = parseInt(timeRangeModel.get(this.options.modelAttribute), 10) || 0;
           		var unit = ((timeRange >= YEAR) && (timeRange % YEAR == 0)) ? YEAR:
                           ((timeRange >= MONTH) && (timeRange % MONTH == 0)) ? MONTH:
                           ((timeRange >= WEEK) && (timeRange % WEEK == 0)) ? WEEK:
                           ((timeRange >= DAY) && (timeRange % DAY == 0)) ? DAY:
                           ((timeRange >= HOUR) && (timeRange % HOUR == 0)) ? HOUR:
                           ((timeRange >= MINUTE) && (timeRange % MINUTE == 0)) ? MINUTE: SECOND;
           		var number = timeRange / unit;
           		timeRangeMediator.set({
           			'number': number,
           			'unit': unit
           		});
            },

            render: function() {

                // TODO: move these styles to CSS file
                // style the input box, to make sure it looks like 'input-append' style.
                this.$el.addClass('input-append').css('margin-bottom', '0px');
                this.children.number.$el.css('float', 'left');
                this.children.unit.$el.css('float', 'right');

            	this.$el.append(this.children.number.render().el);
            	this.$el.append(this.children.unit.render().el);
            	return this;
            }
        });
	}
);